from pathlib import Path
from typing import List, Optional, Dict, Any

from phidata.app import PhidataApp
from phidata.app.databox import default_databox_name
from phidata.infra.base import InfraConfig
from phidata.infra.k8s.args import K8sArgs
from phidata.infra.k8s.manager import K8sManager
from phidata.infra.k8s.resource.group import K8sResourceGroup
from phidata.infra.k8s.create.group import CreateK8sResourceGroup
from phidata.utils.log import logger


class K8sConfig(InfraConfig):
    def __init__(
        self,
        env: Optional[str] = "prd",
        version: Optional[str] = None,
        enabled: bool = True,
        # K8s namespace to use
        namespace: str = "default",
        # K8s context to use
        context: Optional[str] = None,
        # K8s service account to use
        service_account_name: Optional[str] = None,
        # Common K8s labels to add to all resources
        common_labels: Optional[Dict[str, str]] = None,
        # PhidataApp to deploy
        apps: Optional[List[PhidataApp]] = None,
        # K8sResourceGroup to deploy
        resources: Optional[List[K8sResourceGroup]] = None,
        # CreateK8sResourceGroup to deploy
        create_resources: Optional[List[CreateK8sResourceGroup]] = None,
        # Resources dir where k8s manifests are stored
        resources_dir: str = "k8s",
        # databox name for `phi dx ...` commands
        databox_name: str = default_databox_name,
        kubeconfig: Optional[str] = None,
        # Get K8s context and kubeconfig from an EksCluster resource
        eks_cluster: Optional[Any] = None,
        # Aws params for this Config
        # Override the aws params from WorkspaceConfig if provided
        aws_region: Optional[str] = None,
        aws_profile: Optional[str] = None,
        aws_config_file: Optional[str] = None,
        aws_shared_credentials_file: Optional[str] = None,
    ):
        super().__init__()
        try:
            k8s_context = context
            k8s_kubeconfig_path = Path(kubeconfig) if kubeconfig is not None else None
            if eks_cluster is not None:
                from phidata.infra.aws.resource.eks.cluster import EksCluster

                if not isinstance(eks_cluster, EksCluster):
                    raise TypeError("eks_cluster not of type EksCluster")

                k8s_context = eks_cluster.get_kubeconfig_context_name()
                k8s_kubeconfig_path = eks_cluster.kubeconfig_path

            self.args: K8sArgs = K8sArgs(
                env=env,
                version=version,
                enabled=enabled,
                namespace=namespace,
                context=k8s_context,
                service_account_name=service_account_name,
                common_labels=common_labels,
                apps=apps,
                resources=resources,
                create_resources=create_resources,
                databox_name=databox_name,
                resources_dir=resources_dir,
                kubeconfig_path=k8s_kubeconfig_path,
                eks_cluster=eks_cluster,
                aws_region=aws_region,
                aws_profile=aws_profile,
                aws_config_file=aws_config_file,
                aws_shared_credentials_file=aws_shared_credentials_file,
            )
        except Exception as e:
            logger.error(f"Args for {self.__class__.__name__} are not valid")
            raise

    @property
    def namespace(self) -> Optional[str]:
        return self.args.namespace if self.args else None

    @namespace.setter
    def namespace(self, namespace: str) -> None:
        if self.args is not None and namespace is not None:
            self.args.namespace = namespace

    @property
    def context(self) -> Optional[str]:
        return self.args.context if self.args else None

    @context.setter
    def context(self, context: str) -> None:
        if self.args is not None and context is not None:
            self.args.context = context

    @property
    def service_account_name(self) -> Optional[str]:
        return self.args.service_account_name if self.args else None

    @service_account_name.setter
    def service_account_name(self, service_account_name: str) -> None:
        if self.args is not None and service_account_name is not None:
            self.args.service_account_name = service_account_name

    @property
    def common_labels(self) -> Optional[Dict[str, str]]:
        return self.args.common_labels if self.args else None

    @common_labels.setter
    def common_labels(self, common_labels: Dict[str, str]) -> None:
        if self.args is not None and common_labels is not None:
            self.args.common_labels = common_labels

    @property
    def apps(self) -> Optional[List[PhidataApp]]:
        if self.args and self.args.apps:
            return self.args.apps
        return None

    @property
    def resources(self) -> Optional[List[K8sResourceGroup]]:
        if self.args and self.args.resources:
            return self.args.resources
        return None

    @property
    def create_resources(self) -> Optional[List[CreateK8sResourceGroup]]:
        if self.args and self.args.create_resources:
            return self.args.create_resources
        return None

    @property
    def resources_dir(self) -> Optional[str]:
        if self.args and self.args.resources_dir:
            return self.args.resources_dir
        return None

    @property
    def databox_name(self) -> Optional[str]:
        if self.args and self.args.databox_name:
            return self.args.databox_name
        return None

    @property
    def kubeconfig_path(self) -> Optional[Path]:
        if self.args and self.args.kubeconfig_path:
            return self.args.kubeconfig_path
        return None

    @property
    def eks_cluster(self) -> Optional[Any]:
        if self.args and self.args.eks_cluster:
            return self.args.eks_cluster
        return None

    def apps_are_valid(self) -> bool:
        if self.apps is None:
            return True
        for _app in self.apps:
            if not isinstance(_app, PhidataApp):
                raise TypeError("Invalid App: {}".format(_app))
        return True

    def resources_are_valid(self) -> bool:
        if self.resources is None:
            return True
        for _resource in self.resources:
            if not isinstance(_resource, K8sResourceGroup):
                raise TypeError("Invalid K8sResourceGroup: {}".format(_resource))
        return True

    def create_resources_are_valid(self) -> bool:
        if self.create_resources is None:
            return True
        for _resource in self.create_resources:
            if not isinstance(_resource, CreateK8sResourceGroup):
                raise TypeError("Invalid CreateK8sResourceGroup: {}".format(_resource))
        return True

    def is_valid(self) -> bool:
        return (
            self.apps_are_valid()
            and self.resources_are_valid()
            and self.create_resources_are_valid()
        )

    def get_k8s_manager(self) -> Optional[K8sManager]:
        return K8sManager(k8s_args=self.args)

    def get_app_by_name(self, app_name: str) -> Optional[PhidataApp]:

        if self.apps is None:
            return None

        for _app in self.apps:
            try:
                if app_name == _app.name:
                    return _app
            except Exception:
                continue
        return None



import time
import functools

from phidata.utils.log import logger


def timer(func):
    """Print the runtime of the decorated function
    https://realpython.com/primer-on-python-decorators/#timing-functions
    """

    @functools.wraps(func)
    def wrapper_timer(*args, **kwargs):
        start_time = time.perf_counter()  # 1
        value = func(*args, **kwargs)
        end_time = time.perf_counter()  # 2
        run_time = end_time - start_time  # 3
        logger.debug(f"Finished {func.__name__!r} in {run_time:.4f} secs")
        return value

    return wrapper_timer



from typing import Optional, Any, Union, List, Dict

from sqlalchemy.engine import Engine, Connection

from phidata.app.db import DbApp
from phidata.asset.table.sql import SqlTable, SqlTableArgs, SqlType
from phidata.task import Task
from phidata.check import Check
from phidata.utils.log import logger


class PostgresTable(SqlTable):
    def __init__(
        self,
        # Table Name,
        name: Optional[str] = None,
        # Table schema,
        db_schema: Optional[str] = None,
        # sqlalchemy.engine.(Engine or Connection),
        # Using SQLAlchemy makes it possible to use any DB supported by that library.,
        # NOTE: db_engine is required but can be derived using other args.,
        db_engine: Optional[Union[Engine, Connection]] = None,
        # a db_conn_url can be used to create the sqlalchemy.engine.Engine object,
        db_conn_url: Optional[str] = None,
        # airflow connection_id used for running workflows on airflow,
        airflow_conn_id: Optional[str] = None,
        # Phidata DbApp to connect to the database,
        db_app: Optional[DbApp] = None,
        # Checks to run before loading the table,
        pre_checks: Optional[List[Check]] = None,
        # List of tasks to create the table
        create_tasks: Optional[List[Task]] = None,
        # Checks to run after loading the table,
        post_checks: Optional[List[Check]] = None,
        # List of tasks to update the table,
        update_tasks: Optional[List[Task]] = None,
        # List of tasks to delete the table,
        delete_tasks: Optional[List[Task]] = None,
        # Control which environment the table is created in
        env: Optional[str] = None,
        # Dev Args,
        dev_name: Optional[str] = None,
        dev_env: Optional[Dict[str, Any]] = None,
        seed_dev_tasks: Optional[List[Task]] = None,
        dev_stg_swap_tasks: Optional[List[Task]] = None,
        # Staging Args,
        stg_name: Optional[str] = None,
        stg_env: Optional[Dict[str, Any]] = None,
        seed_stg_tasks: Optional[List[Task]] = None,
        stg_prd_swap_tasks: Optional[List[Task]] = None,
        # Production Args,
        prd_name: Optional[str] = None,
        prd_env: Optional[Dict[str, Any]] = None,
        cached_db_engine: Optional[Union[Engine, Connection]] = None,
        version: Optional[str] = None,
        enabled: bool = True,
        **kwargs,
    ) -> None:
        super().__init__()
        try:
            self.args: SqlTableArgs = SqlTableArgs(
                name=name,
                sql_type=SqlType.POSTGRES,
                db_schema=db_schema,
                db_engine=db_engine,
                db_conn_url=db_conn_url,
                airflow_conn_id=airflow_conn_id,
                db_app=db_app,
                pre_checks=pre_checks,
                create_tasks=create_tasks,
                post_checks=post_checks,
                update_tasks=update_tasks,
                delete_tasks=delete_tasks,
                env=env,
                dev_name=dev_name,
                dev_env=dev_env,
                seed_dev_tasks=seed_dev_tasks,
                dev_stg_swap_tasks=dev_stg_swap_tasks,
                stg_name=stg_name,
                stg_env=stg_env,
                seed_stg_tasks=seed_stg_tasks,
                stg_prd_swap_tasks=stg_prd_swap_tasks,
                prd_name=prd_name,
                prd_env=prd_env,
                cached_db_engine=cached_db_engine,
                version=version,
                enabled=enabled,
                **kwargs,
            )
        except Exception as e:
            logger.error(f"Args for {self.name} are not valid")
            raise



from typing import Optional

from pydantic import Field
from kubernetes.client.models.v1_toleration import V1Toleration

from phidata.infra.k8s.resource.base import K8sObject


class Toleration(K8sObject):
    """
    The pod this Toleration is attached to tolerates any taint that matches
    the triple <key,value,effect> using the matching operator <operator>.

    # https://kubernetes.io/docs/reference/generated/kubernetes-api/v1.23/#toleration-v1-core
    """

    resource_type: str = "Toleration"

    # Effect indicates the taint effect to match.
    # Empty means match all taint effects.
    # When specified, allowed values are NoSchedule, PreferNoSchedule and NoExecute.
    effect: Optional[str] = None
    # Key is the taint key that the toleration applies to. Empty means match all taint keys.
    # If the key is empty, operator must be Exists; this combination means to match all values and all keys.
    key: Optional[str] = None
    # Operator represents a key's relationship to the value. Valid operators are Exists and Equal. Defaults to Equal.
    # Exists is equivalent to wildcard for value, so that a pod can tolerate all taints of a particular category.
    # Possible enum values: - `"Equal"` - `"Exists"`
    operator: Optional[str] = None
    # TolerationSeconds represents the period of time the toleration (which must be of effect NoExecute,
    # otherwise this field is ignored) tolerates the taint. By default, it is not set, which means tolerate the
    # taint forever (do not evict). Zero and negative values will be treated as 0 (evict immediately) by the system.
    toleration_seconds: Optional[int] = Field(None, alias="tolerationSeconds")
    # Value is the taint value the toleration matches to. If the operator is Exists, the value should be empty,
    # otherwise just a regular string.
    value: Optional[str] = None

    def get_k8s_object(self) -> V1Toleration:

        # Return a V1Toleration object
        # https://github.com/kubernetes-client/python/blob/master/kubernetes/client/models/v1_toleration.py
        _v1_toleration = V1Toleration(
            effect=self.effect,
            key=self.key,
            operator=self.operator,
            toleration_seconds=self.toleration_seconds,
            value=self.value,
        )
        return _v1_toleration



from collections import OrderedDict
from pathlib import Path
from typing import Optional, Dict, Any, List, Union
from typing_extensions import Literal

from phidata.app.db import DbApp
from phidata.app import PhidataApp, PhidataAppArgs
from phidata.constants import (
    SCRIPTS_DIR_ENV_VAR,
    STORAGE_DIR_ENV_VAR,
    META_DIR_ENV_VAR,
    PRODUCTS_DIR_ENV_VAR,
    NOTEBOOKS_DIR_ENV_VAR,
    WORKSPACE_CONFIG_DIR_ENV_VAR,
    PHIDATA_RUNTIME_ENV_VAR,
)
from phidata.infra.docker.resource.group import (
    DockerResourceGroup,
    DockerBuildContext,
    DockerContainer,
    DockerNetwork,
)
from phidata.infra.k8s.create.apps.v1.deployment import RestartPolicy
from phidata.infra.k8s.create.core.v1.service import ServiceType
from phidata.infra.k8s.create.core.v1.container import CreateContainer, ImagePullPolicy
from phidata.infra.k8s.create.core.v1.persistent_volume import (
    ClaimRef,
    NFSVolumeSource,
)
from phidata.infra.k8s.create.rbac_authorization_k8s_io.v1.cluster_role import (
    PolicyRule,
)
from phidata.infra.k8s.create.core.v1.volume import (
    CreateVolume,
    HostPathVolumeSource,
    VolumeType,
    PersistentVolumeClaimVolumeSource,
)
from phidata.infra.k8s.create.common.port import CreatePort
from phidata.infra.k8s.create.group import (
    CreateK8sResourceGroup,
    CreateNamespace,
    CreateServiceAccount,
    CreateClusterRole,
    CreateClusterRoleBinding,
    CreateSecret,
    CreateConfigMap,
    CreateStorageClass,
    CreateService,
    CreateDeployment,
    CreateCustomObject,
    CreateCustomResourceDefinition,
    CreatePersistentVolume,
    CreatePVC,
)
from phidata.infra.k8s.enums.pv import PVAccessMode
from phidata.infra.k8s.resource.group import (
    K8sResourceGroup,
    K8sBuildContext,
)
from phidata.utils.common import (
    get_image_str,
    get_default_ns_name,
    get_default_container_name,
    get_default_configmap_name,
    get_default_secret_name,
    get_default_service_name,
    get_default_deploy_name,
    get_default_pod_name,
    get_default_volume_name,
    get_default_cr_name,
    get_default_crb_name,
    get_default_sa_name,
)
from phidata.utils.enums import ExtendedEnum
from phidata.utils.log import logger


class SupersetBaseArgs(PhidataAppArgs):
    name: str = "superset"
    version: str = "1"
    enabled: bool = True

    # Image args
    image_name: str = "phidata/superset"
    image_tag: str = "2.0.0"
    entrypoint: Optional[Union[str, List]] = None
    command: Optional[Union[str, List]] = None

    # Install python dependencies using a requirements.txt file
    # Sets the REQUIREMENTS_LOCAL & REQUIREMENTS_FILE_PATH env var to requirements_file_path
    install_requirements: bool = False
    # Path to the requirements.txt file relative to the workspace_root
    requirements_file_path: str = "requirements.txt"

    # Configure Superset db
    wait_for_db: bool = False
    # Connect to database using a DbApp
    db_app: Optional[DbApp] = None
    # Provide database connection details manually
    # db_user can be provided here or as the
    # DATABASE_USER env var in the secrets_file
    db_user: Optional[str] = None
    # db_password can be provided here or as the
    # DATABASE_PASSWORD env var in the secrets_file
    db_password: Optional[str] = None
    # db_schema can be provided here or as the
    # DATABASE_DB env var in the secrets_file
    db_schema: Optional[str] = None
    # db_host can be provided here or as the
    # DATABASE_HOST env var in the secrets_file
    db_host: Optional[str] = None
    # db_port can be provided here or as the
    # DATABASE_PORT env var in the secrets_file
    db_port: Optional[int] = None
    # db_driver can be provided here or as the
    # DATABASE_DIALECT env var in the secrets_file
    db_dialect: Optional[str] = None

    # Configure superset redis
    wait_for_redis: bool = False
    # Connect to redis using a PhidataApp
    redis_app: Optional[DbApp] = None
    # redis_host can be provided here or as the
    # REDIS_HOST env var in the secrets_file
    redis_host: Optional[str] = None
    # redis_port can be provided here or as the
    # REDIS_PORT env var in the secrets_file
    redis_port: Optional[int] = None
    # redis_driver can be provided here or as the
    # REDIS_DRIVER env var in the secrets_file
    redis_driver: Optional[str] = None

    # Configure the superset container
    container_name: Optional[str] = None
    # Set the SUPERSET_CONFIG_PATH env var
    superset_config_path: Optional[str] = None
    # Set the FLASK_ENV env var
    flask_env: str = "production"
    # Set the SUPERSET_ENV env var
    superset_env: str = "production"
    # Set the PYTHONPATH env var
    # defaults to "/app/pythonpath:/app/docker/pythonpath_dev"
    python_path: Optional[str] = None
    # Add labels to the container
    container_labels: Optional[Dict[str, Any]] = None

    # Docker configuration
    # NOTE: Only available for Docker
    # Run container in the background and return a Container object.
    container_detach: bool = True
    # Enable auto-removal of the container on daemon side when the container’s process exits.
    container_auto_remove: bool = True
    # Remove the container when it has finished running. Default: False.
    container_remove: bool = True
    # Username or UID to run commands as inside the container.
    container_user: Optional[Union[str, int]] = None
    # Keep STDIN open even if not attached.
    container_stdin_open: bool = True
    container_tty: bool = True
    # Specify a test to perform to check that the container is healthy.
    container_healthcheck: Optional[Dict[str, Any]] = None
    # Optional hostname for the container.
    container_hostname: Optional[str] = None
    # Platform in the format os[/arch[/variant]].
    container_platform: Optional[str] = None
    # Path to the working directory.
    container_working_dir: Optional[str] = None
    # Restart the container when it exits. Configured as a dictionary with keys:
    # Name: One of on-failure, or always.
    # MaximumRetryCount: Number of times to restart the container on failure.
    # For example: {"Name": "on-failure", "MaximumRetryCount": 5}
    container_restart_policy_docker: Optional[Dict[str, Any]] = None
    # Add volumes to DockerContainer
    # container_volumes is a dictionary which adds the volumes to mount
    # inside the container. The key is either the host path or a volume name,
    # and the value is a dictionary with 2 keys:
    #   bind - The path to mount the volume inside the container
    #   mode - Either rw to mount the volume read/write, or ro to mount it read-only.
    # For example:
    # {
    #   '/home/user1/': {'bind': '/mnt/vol2', 'mode': 'rw'},
    #   '/var/www': {'bind': '/mnt/vol1', 'mode': 'ro'}
    # }
    container_volumes_docker: Optional[Dict[str, dict]] = None
    # Add ports to DockerContainer
    # The keys of the dictionary are the ports to bind inside the container,
    # either as an integer or a string in the form port/protocol, where the protocol is either tcp, udp.
    # The values of the dictionary are the corresponding ports to open on the host, which can be either:
    #   - The port number, as an integer.
    #       For example, {'2222/tcp': 3333} will expose port 2222 inside the container as port 3333 on the host.
    #   - None, to assign a random host port. For example, {'2222/tcp': None}.
    #   - A tuple of (address, port) if you want to specify the host interface.
    #       For example, {'1111/tcp': ('127.0.0.1', 1111)}.
    #   - A list of integers, if you want to bind multiple host ports to a single container port.
    #       For example, {'1111/tcp': [1234, 4567]}.
    container_ports_docker: Optional[Dict[str, Any]] = None

    # K8s configuration
    # NOTE: Only available for Kubernetes
    image_pull_policy: ImagePullPolicy = ImagePullPolicy.IF_NOT_PRESENT

    # Container ports
    # Open a container port if open_container_port=True
    open_container_port: bool = False
    # Port number on the container
    container_port: int = 8000
    # Port name: Only used by the K8sContainer
    container_port_name: str = "http"
    # Host port: Only used by the DockerContainer
    container_host_port: int = 8000

    # Open the app port if open_app_port=True
    open_app_port: bool = False
    # App port number on the container
    # Set the SUPERSET_PORT env var
    app_port: int = 8088
    # Only used by the K8sContainer
    app_port_name: str = "app"
    # Only used by the DockerContainer
    app_host_port: int = 8088

    # Container env
    # Add env variables to container env
    env: Optional[Dict[str, str]] = None
    # Read env variables from a file in yaml format
    env_file: Optional[Path] = None
    # Configure the ConfigMap name used for env variables that are not Secret
    config_map_name: Optional[str] = None

    # Container secrets
    # Add secret variables to container env
    secrets: Optional[Dict[str, str]] = None
    # Read secret variables from a file in yaml format
    secrets_file: Optional[Path] = None
    # Read secret variables from AWS Secrets Manager
    aws_secret: Optional[Any] = None
    # Configure the Secret name used for env variables that are Secret
    secret_name: Optional[str] = None

    # Container volumes
    # Configure workspace volume
    # Mount the workspace directory on the container
    mount_workspace: bool = False
    workspace_volume_name: Optional[str] = None
    # Path to mount the workspace volume under
    # This is the parent directory for the workspace on the container
    # i.e. the ws is mounted as a subdir in this dir
    # eg: if ws name is: idata, workspace_root would be: /mnt/workspaces/idata
    workspace_mount_container_path: str = "/mnt/workspaces"
    # NOTE: On DockerContainers the local workspace_root_path is mounted under workspace_mount_container_path
    # because we assume that DockerContainers are running locally on the user's machine
    # On K8sContainers, we load the workspace_dir from git using a git-sync sidecar container
    create_git_sync_sidecar: bool = False
    create_git_sync_init_container: bool = True
    git_sync_repo: Optional[str] = None
    git_sync_branch: Optional[str] = None
    git_sync_wait: int = 1
    # When running k8s locally, we can mount the workspace using host path as well.
    k8s_mount_local_workspace = False

    # Configure resources volume - only on docker
    # Superset resources directory relative to the workspace_root
    # This directory contains all the files required by superset.
    # eg: docker-bootstrap.sh
    # This dir is mounted to the `/app/docker` directory on the container
    mount_resources: bool = False
    resources_dir: str = "workspace/superset"
    resources_dir_container_path: str = "/app/docker"
    resources_volume_name: Optional[str] = None

    # Configure the deployment
    deploy_name: Optional[str] = None
    pod_name: Optional[str] = None
    replicas: int = 1
    pod_annotations: Optional[Dict[str, str]] = None
    pod_node_selector: Optional[Dict[str, str]] = None
    deploy_restart_policy: RestartPolicy = RestartPolicy.ALWAYS
    termination_grace_period_seconds: Optional[int] = None
    # Add deployment labels
    deploy_labels: Optional[Dict[str, Any]] = None
    # Determine how to spread the deployment across a topology
    # Key to spread the pods across
    topology_spread_key: Optional[str] = None
    # The degree to which pods may be unevenly distributed
    topology_spread_max_skew: Optional[int] = None
    # How to deal with a pod if it doesn't satisfy the spread constraint.
    topology_spread_when_unsatisfiable: Optional[
        Literal["DoNotSchedule", "ScheduleAnyway"]
    ] = None

    # Configure the app service
    create_app_service: bool = False
    app_svc_name: Optional[str] = None
    app_svc_type: Optional[ServiceType] = None
    # The port that will be exposed by the service.
    app_svc_port: int = 8088
    # The node_port that will be exposed by the service if app_svc_type = ServiceType.NODE_PORT
    app_node_port: Optional[int] = None
    # The app_target_port is the port to access on the pods targeted by the service.
    # It can be the port number or port name on the pod.
    app_target_port: Optional[Union[str, int]] = None
    # Extra ports exposed by the app service
    app_svc_ports: Optional[List[CreatePort]] = None
    # Add labels to app service
    app_svc_labels: Optional[Dict[str, Any]] = None
    # Add annotations to app service
    app_svc_annotations: Optional[Dict[str, str]] = None
    # If ServiceType == LoadBalancer
    app_svc_health_check_node_port: Optional[int] = None
    app_svc_internal_taffic_policy: Optional[str] = None
    app_svc_load_balancer_class: Optional[str] = None
    app_svc_load_balancer_ip: Optional[str] = None
    app_svc_load_balancer_source_ranges: Optional[List[str]] = None
    app_svc_allocate_load_balancer_node_ports: Optional[bool] = None

    # Configure K8s rbac: use a separate Namespace, ServiceAccount,
    # ClusterRole & ClusterRoleBinding
    use_rbac: bool = False
    # Create a Namespace with name ns_name & default values
    ns_name: Optional[str] = None
    # Provide the full Namespace definition
    namespace: Optional[CreateNamespace] = None
    # Create a ServiceAccount with name sa_name & default values
    sa_name: Optional[str] = None
    # Provide the full ServiceAccount definition
    service_account: Optional[CreateServiceAccount] = None
    # Create a ClusterRole with name sa_name & default values
    cr_name: Optional[str] = None
    # Provide the full ClusterRole definition
    cluster_role: Optional[CreateClusterRole] = None
    # Create a ClusterRoleBinding with name sa_name & default values
    crb_name: Optional[str] = None
    # Provide the full ClusterRoleBinding definition
    cluster_role_binding: Optional[CreateClusterRoleBinding] = None

    # Other args
    # Set SUPERSET_LOAD_EXAMPLES = "yes"
    load_examples: bool = False
    print_env_on_load: bool = True

    # Add extra Kubernetes resources
    extra_secrets: Optional[List[CreateSecret]] = None
    extra_config_maps: Optional[List[CreateConfigMap]] = None
    extra_storage_classes: Optional[List[CreateStorageClass]] = None
    extra_services: Optional[List[CreateService]] = None
    extra_deployments: Optional[List[CreateDeployment]] = None
    extra_custom_objects: Optional[List[CreateCustomObject]] = None
    extra_crds: Optional[List[CreateCustomResourceDefinition]] = None
    extra_pvs: Optional[List[CreatePersistentVolume]] = None
    extra_pvcs: Optional[List[CreatePVC]] = None
    extra_containers: Optional[List[CreateContainer]] = None
    extra_init_containers: Optional[List[CreateContainer]] = None
    extra_ports: Optional[List[CreatePort]] = None
    extra_volumes: Optional[List[CreateVolume]] = None


class SupersetBase(PhidataApp):
    def __init__(
        self,
        name: str = "superset",
        version: str = "1",
        enabled: bool = True,
        # Image args,
        image_name: str = "phidata/superset",
        image_tag: str = "2.0.0",
        entrypoint: Optional[Union[str, List]] = None,
        command: Optional[Union[str, List]] = None,
        # Install python dependencies using a requirements.txt file,
        # Sets the REQUIREMENTS_LOCAL & REQUIREMENTS_FILE_PATH env var to requirements_file_path,
        install_requirements: bool = False,
        # Path to the requirements.txt file relative to the workspace_root,
        requirements_file_path: str = "requirements.txt",
        # Configure Superset db,
        wait_for_db: bool = False,
        # Connect to database using a DbApp,
        db_app: Optional[DbApp] = None,
        # Provide database connection details manually,
        # db_user can be provided here or as the,
        # DATABASE_USER env var in the secrets_file,
        db_user: Optional[str] = None,
        # db_password can be provided here or as the,
        # DATABASE_PASSWORD env var in the secrets_file,
        db_password: Optional[str] = None,
        # db_schema can be provided here or as the,
        # DATABASE_DB env var in the secrets_file,
        db_schema: Optional[str] = None,
        # db_host can be provided here or as the,
        # DATABASE_HOST env var in the secrets_file,
        db_host: Optional[str] = None,
        # db_port can be provided here or as the,
        # DATABASE_PORT env var in the secrets_file,
        db_port: Optional[int] = None,
        # db_driver can be provided here or as the,
        # DATABASE_DIALECT env var in the secrets_file,
        db_dialect: Optional[str] = None,
        # Configure superset redis,
        wait_for_redis: bool = False,
        # Connect to redis using a PhidataApp,
        redis_app: Optional[DbApp] = None,
        # redis_host can be provided here or as the,
        # REDIS_HOST env var in the secrets_file,
        redis_host: Optional[str] = None,
        # redis_port can be provided here or as the,
        # REDIS_PORT env var in the secrets_file,
        redis_port: Optional[int] = None,
        # redis_driver can be provided here or as the,
        # REDIS_DRIVER env var in the secrets_file,
        redis_driver: Optional[str] = None,
        # Configure the superset container,
        container_name: Optional[str] = None,
        # Set the SUPERSET_CONFIG_PATH env var,
        superset_config_path: Optional[str] = None,
        # Set the FLASK_ENV env var,
        flask_env: str = "production",
        # Set the SUPERSET_ENV env var,
        superset_env: str = "production",
        # Set the PYTHONPATH env var,
        # defaults to "/app/pythonpath:/app/docker/pythonpath_dev",
        python_path: Optional[str] = None,
        # Add labels to the container,
        container_labels: Optional[Dict[str, Any]] = None,
        # Docker configuration,
        # NOTE: Only available for Docker,
        # Run container in the background and return a Container object.,
        container_detach: bool = True,
        # Enable auto-removal of the container on daemon side when the container’s process exits.,
        container_auto_remove: bool = True,
        # Remove the container when it has finished running. Default: False.,
        container_remove: bool = True,
        # Username or UID to run commands as inside the container.,
        container_user: Optional[Union[str, int]] = None,
        # Keep STDIN open even if not attached.,
        container_stdin_open: bool = True,
        container_tty: bool = True,
        # Specify a test to perform to check that the container is healthy.,
        container_healthcheck: Optional[Dict[str, Any]] = None,
        # Optional hostname for the container.,
        container_hostname: Optional[str] = None,
        # Platform in the format os[/arch[/variant]].,
        container_platform: Optional[str] = None,
        # Path to the working directory.,
        container_working_dir: Optional[str] = None,
        # Restart the container when it exits. Configured as a dictionary with keys:,
        # Name: One of on-failure, or always.,
        # MaximumRetryCount: Number of times to restart the container on failure.,
        # For example: {"Name": "on-failure", "MaximumRetryCount": 5},
        container_restart_policy_docker: Optional[Dict[str, Any]] = None,
        # Add volumes to DockerContainer,
        # container_volumes is a dictionary which adds the volumes to mount,
        # inside the container. The key is either the host path or a volume name,,
        # and the value is a dictionary with 2 keys:,
        #   bind - The path to mount the volume inside the container,
        #   mode - Either rw to mount the volume read/write, or ro to mount it read-only.,
        # For example:,
        # {,
        #   '/home/user1/': {'bind': '/mnt/vol2', 'mode': 'rw'},,
        #   '/var/www': {'bind': '/mnt/vol1', 'mode': 'ro'},
        # },
        container_volumes_docker: Optional[Dict[str, dict]] = None,
        # Add ports to DockerContainer,
        # The keys of the dictionary are the ports to bind inside the container,,
        # either as an integer or a string in the form port/protocol, where the protocol is either tcp, udp.,
        # The values of the dictionary are the corresponding ports to open on the host, which can be either:,
        #   - The port number, as an integer.,
        #       For example, {'2222/tcp': 3333} will expose port 2222 inside the container as port 3333 on the host.,
        #   - None, to assign a random host port. For example, {'2222/tcp': None}.,
        #   - A tuple of (address, port) if you want to specify the host interface.,
        #       For example, {'1111/tcp': ('127.0.0.1', 1111)}.,
        #   - A list of integers, if you want to bind multiple host ports to a single container port.,
        #       For example, {'1111/tcp': [1234, 4567]}.,
        container_ports_docker: Optional[Dict[str, Any]] = None,
        # K8s configuration,
        # NOTE: Only available for Kubernetes,
        image_pull_policy: ImagePullPolicy = ImagePullPolicy.IF_NOT_PRESENT,
        # Container ports,
        # Open a container port if open_container_port=True,
        open_container_port: bool = False,
        # Port number on the container,
        container_port: int = 8000,
        # Port name: Only used by the K8sContainer,
        container_port_name: str = "http",
        # Host port: Only used by the DockerContainer,
        container_host_port: int = 8000,
        # Open the app port if open_app_port=True,
        open_app_port: bool = False,
        # App port number on the container,
        # Set the SUPERSET_PORT env var,
        app_port: int = 8088,
        # Only used by the K8sContainer,
        app_port_name: str = "app",
        # Only used by the DockerContainer,
        app_host_port: int = 8088,
        # Container env,
        # Add env variables to container env,
        env: Optional[Dict[str, str]] = None,
        # Read env variables from a file in yaml format,
        env_file: Optional[Path] = None,
        # Configure the ConfigMap name used for env variables that are not Secret,
        config_map_name: Optional[str] = None,
        # Container secrets,
        # Add secret variables to container env,
        secrets: Optional[Dict[str, str]] = None,
        # Read secret variables from a file in yaml format,
        secrets_file: Optional[Path] = None,
        # Read secret variables from AWS Secrets Manager,
        aws_secret: Optional[Any] = None,
        # Configure the Secret name used for env variables that are Secret,
        secret_name: Optional[str] = None,
        # Container volumes,
        # Configure workspace volume,
        # Mount the workspace directory on the container,
        mount_workspace: bool = False,
        workspace_volume_name: Optional[str] = None,
        # Path to mount the workspace volume under,
        # This is the parent directory for the workspace on the container,
        # i.e. the ws is mounted as a subdir in this dir,
        # eg: if ws name is: idata, workspace_dir would be: /workspaces/idata,
        # eg: if ws name is: idata, workspace_root would be: /mnt/workspaces/idata,
        workspace_mount_container_path: str = "/mnt/workspaces",
        # NOTE: On DockerContainers the local workspace_root_path is mounted under workspace_mount_container_path,
        # because we assume that DockerContainers are running locally on the user's machine,
        # On K8sContainers, we load the workspace_dir from git using a git-sync sidecar container,
        create_git_sync_sidecar: bool = False,
        create_git_sync_init_container: bool = True,
        git_sync_repo: Optional[str] = None,
        git_sync_branch: Optional[str] = None,
        git_sync_wait: int = 1,
        # When running k8s locally, we can mount the workspace using host path as well.,
        k8s_mount_local_workspace=False,
        # Configure resources volume - only on docker,
        # Superset resources directory relative to the workspace_root,
        # This directory contains all the files required by superset.,
        # eg: docker-bootstrap.sh,
        # This dir is mounted to the `/app/docker` directory on the container,
        mount_resources: bool = False,
        resources_dir: str = "workspace/superset",
        resources_dir_container_path: str = "/app/docker",
        resources_volume_name: Optional[str] = None,
        # Configure the deployment,
        deploy_name: Optional[str] = None,
        pod_name: Optional[str] = None,
        replicas: int = 1,
        pod_annotations: Optional[Dict[str, str]] = None,
        pod_node_selector: Optional[Dict[str, str]] = None,
        deploy_restart_policy: RestartPolicy = RestartPolicy.ALWAYS,
        termination_grace_period_seconds: Optional[int] = None,
        # Add deployment labels,
        deploy_labels: Optional[Dict[str, Any]] = None,
        # Determine how to spread the deployment across a topology,
        # Key to spread the pods across,
        topology_spread_key: Optional[str] = None,
        # The degree to which pods may be unevenly distributed,
        topology_spread_max_skew: Optional[int] = None,
        # How to deal with a pod if it doesn't satisfy the spread constraint.,
        topology_spread_when_unsatisfiable: Optional[
            Literal["DoNotSchedule", "ScheduleAnyway"]
        ] = None,
        # Configure the app service,
        create_app_service: bool = False,
        app_svc_name: Optional[str] = None,
        app_svc_type: Optional[ServiceType] = None,
        # The port that will be exposed by the service.,
        app_svc_port: int = 8088,
        # The node_port that will be exposed by the service if app_svc_type = ServiceType.NODE_PORT,
        app_node_port: Optional[int] = None,
        # The app_target_port is the port to access on the pods targeted by the service.,
        # It can be the port number or port name on the pod.,
        app_target_port: Optional[Union[str, int]] = None,
        # Extra ports exposed by the app service,
        app_svc_ports: Optional[List[CreatePort]] = None,
        # Add labels to app service,
        app_svc_labels: Optional[Dict[str, Any]] = None,
        # Add annotations to app service,
        app_svc_annotations: Optional[Dict[str, str]] = None,
        # If ServiceType == LoadBalancer,
        app_svc_health_check_node_port: Optional[int] = None,
        app_svc_internal_taffic_policy: Optional[str] = None,
        app_svc_load_balancer_class: Optional[str] = None,
        app_svc_load_balancer_ip: Optional[str] = None,
        app_svc_load_balancer_source_ranges: Optional[List[str]] = None,
        app_svc_allocate_load_balancer_node_ports: Optional[bool] = None,
        # Configure K8s rbac: use a separate Namespace, ServiceAccount,,
        # ClusterRole & ClusterRoleBinding,
        use_rbac: bool = False,
        # Create a Namespace with name ns_name & default values,
        ns_name: Optional[str] = None,
        # Provide the full Namespace definition,
        namespace: Optional[CreateNamespace] = None,
        # Create a ServiceAccount with name sa_name & default values,
        sa_name: Optional[str] = None,
        # Provide the full ServiceAccount definition,
        service_account: Optional[CreateServiceAccount] = None,
        # Create a ClusterRole with name sa_name & default values,
        cr_name: Optional[str] = None,
        # Provide the full ClusterRole definition,
        cluster_role: Optional[CreateClusterRole] = None,
        # Create a ClusterRoleBinding with name sa_name & default values,
        crb_name: Optional[str] = None,
        # Provide the full ClusterRoleBinding definition,
        cluster_role_binding: Optional[CreateClusterRoleBinding] = None,
        # Other args,
        # Set SUPERSET_LOAD_EXAMPLES = "yes",
        load_examples: bool = False,
        print_env_on_load: bool = True,
        # Add extra Kubernetes resources,
        extra_secrets: Optional[List[CreateSecret]] = None,
        extra_config_maps: Optional[List[CreateConfigMap]] = None,
        extra_storage_classes: Optional[List[CreateStorageClass]] = None,
        extra_services: Optional[List[CreateService]] = None,
        extra_deployments: Optional[List[CreateDeployment]] = None,
        extra_custom_objects: Optional[List[CreateCustomObject]] = None,
        extra_crds: Optional[List[CreateCustomResourceDefinition]] = None,
        extra_pvs: Optional[List[CreatePersistentVolume]] = None,
        extra_pvcs: Optional[List[CreatePVC]] = None,
        extra_containers: Optional[List[CreateContainer]] = None,
        extra_init_containers: Optional[List[CreateContainer]] = None,
        extra_ports: Optional[List[CreatePort]] = None,
        extra_volumes: Optional[List[CreateVolume]] = None,
        # If True, use cached resources
        # i.e. skip resource creation/deletion if active resources with the same name exist.
        use_cache: bool = True,
        **extra_kwargs,
    ):
        super().__init__()

        # Cache env_data & secret_data
        self.env_data: Optional[Dict[str, Any]] = None
        self.secret_data: Optional[Dict[str, Any]] = None

        try:
            self.args: SupersetBaseArgs = SupersetBaseArgs(
                name=name,
                version=version,
                enabled=enabled,
                image_name=image_name,
                image_tag=image_tag,
                entrypoint=entrypoint,
                command=command,
                install_requirements=install_requirements,
                requirements_file_path=requirements_file_path,
                wait_for_db=wait_for_db,
                db_app=db_app,
                db_user=db_user,
                db_password=db_password,
                db_schema=db_schema,
                db_host=db_host,
                db_port=db_port,
                db_dialect=db_dialect,
                wait_for_redis=wait_for_redis,
                redis_app=redis_app,
                redis_host=redis_host,
                redis_port=redis_port,
                redis_driver=redis_driver,
                container_name=container_name,
                superset_config_path=superset_config_path,
                flask_env=flask_env,
                superset_env=superset_env,
                python_path=python_path,
                container_labels=container_labels,
                container_detach=container_detach,
                container_auto_remove=container_auto_remove,
                container_remove=container_remove,
                container_user=container_user,
                container_stdin_open=container_stdin_open,
                container_tty=container_tty,
                container_healthcheck=container_healthcheck,
                container_hostname=container_hostname,
                container_platform=container_platform,
                container_working_dir=container_working_dir,
                container_restart_policy_docker=container_restart_policy_docker,
                container_volumes_docker=container_volumes_docker,
                container_ports_docker=container_ports_docker,
                image_pull_policy=image_pull_policy,
                open_container_port=open_container_port,
                container_port=container_port,
                container_port_name=container_port_name,
                container_host_port=container_host_port,
                open_app_port=open_app_port,
                app_port=app_port,
                app_port_name=app_port_name,
                app_host_port=app_host_port,
                env=env,
                env_file=env_file,
                config_map_name=config_map_name,
                secrets=secrets,
                secrets_file=secrets_file,
                aws_secret=aws_secret,
                secret_name=secret_name,
                mount_workspace=mount_workspace,
                workspace_volume_name=workspace_volume_name,
                workspace_mount_container_path=workspace_mount_container_path,
                create_git_sync_sidecar=create_git_sync_sidecar,
                create_git_sync_init_container=create_git_sync_init_container,
                git_sync_repo=git_sync_repo,
                git_sync_branch=git_sync_branch,
                git_sync_wait=git_sync_wait,
                k8s_mount_local_workspace=k8s_mount_local_workspace,
                mount_resources=mount_resources,
                resources_dir=resources_dir,
                resources_dir_container_path=resources_dir_container_path,
                resources_volume_name=resources_volume_name,
                deploy_name=deploy_name,
                pod_name=pod_name,
                replicas=replicas,
                pod_annotations=pod_annotations,
                pod_node_selector=pod_node_selector,
                deploy_restart_policy=deploy_restart_policy,
                termination_grace_period_seconds=termination_grace_period_seconds,
                deploy_labels=deploy_labels,
                topology_spread_key=topology_spread_key,
                topology_spread_max_skew=topology_spread_max_skew,
                topology_spread_when_unsatisfiable=topology_spread_when_unsatisfiable,
                create_app_service=create_app_service,
                app_svc_name=app_svc_name,
                app_svc_type=app_svc_type,
                app_svc_port=app_svc_port,
                app_node_port=app_node_port,
                app_target_port=app_target_port,
                app_svc_ports=app_svc_ports,
                app_svc_labels=app_svc_labels,
                app_svc_annotations=app_svc_annotations,
                app_svc_health_check_node_port=app_svc_health_check_node_port,
                app_svc_internal_taffic_policy=app_svc_internal_taffic_policy,
                app_svc_load_balancer_class=app_svc_load_balancer_class,
                app_svc_load_balancer_ip=app_svc_load_balancer_ip,
                app_svc_load_balancer_source_ranges=app_svc_load_balancer_source_ranges,
                app_svc_allocate_load_balancer_node_ports=app_svc_allocate_load_balancer_node_ports,
                use_rbac=use_rbac,
                ns_name=ns_name,
                namespace=namespace,
                sa_name=sa_name,
                service_account=service_account,
                cr_name=cr_name,
                cluster_role=cluster_role,
                crb_name=crb_name,
                cluster_role_binding=cluster_role_binding,
                load_examples=load_examples,
                print_env_on_load=print_env_on_load,
                extra_secrets=extra_secrets,
                extra_config_maps=extra_config_maps,
                extra_storage_classes=extra_storage_classes,
                extra_services=extra_services,
                extra_deployments=extra_deployments,
                extra_custom_objects=extra_custom_objects,
                extra_crds=extra_crds,
                extra_pvs=extra_pvs,
                extra_pvcs=extra_pvcs,
                extra_containers=extra_containers,
                extra_init_containers=extra_init_containers,
                extra_ports=extra_ports,
                extra_volumes=extra_volumes,
                use_cache=use_cache,
                extra_kwargs=extra_kwargs,
            )
        except Exception as e:
            logger.error(f"Args for {self.__class__.__name__} are not valid")
            raise

    def get_container_name(self) -> str:
        return self.args.container_name or get_default_container_name(self.args.name)

    def get_app_service_name(self) -> str:
        return self.args.app_svc_name or get_default_service_name(self.args.name)

    def get_app_service_port(self) -> int:
        return self.args.app_svc_port

    def get_env_data(self) -> Optional[Dict[str, str]]:
        if self.env_data is None:
            self.env_data = self.read_yaml_file(file_path=self.args.env_file)
        return self.env_data

    def get_secret_data(self) -> Optional[Dict[str, str]]:
        if self.secret_data is None:
            self.secret_data = self.read_yaml_file(file_path=self.args.secrets_file)
        return self.secret_data

    def get_db_user(self) -> Optional[str]:
        db_user_var: Optional[str] = self.args.db_user if self.args else None
        if db_user_var is None:
            # read from secrets_file
            logger.debug(f"Reading DATABASE_USER from secrets")
            secret_data = self.get_secret_data()
            if secret_data is not None:
                db_user_var = secret_data.get("DATABASE_USER", db_user_var)
        return db_user_var

    def get_db_password(self) -> Optional[str]:
        db_password_var: Optional[str] = self.args.db_password if self.args else None
        if db_password_var is None:
            # read from secrets_file
            logger.debug(f"Reading DATABASE_PASSWORD from secrets")
            secret_data = self.get_secret_data()
            if secret_data is not None:
                db_password_var = secret_data.get("DATABASE_PASSWORD", db_password_var)
        return db_password_var

    def get_db_schema(self) -> Optional[str]:
        db_schema_var: Optional[str] = self.args.db_schema if self.args else None
        if db_schema_var is None:
            # read from secrets_file
            logger.debug(f"Reading DATABASE_DB from secrets")
            secret_data = self.get_secret_data()
            if secret_data is not None:
                db_schema_var = secret_data.get("DATABASE_DB", db_schema_var)
        return db_schema_var

    def get_db_host(self) -> Optional[str]:
        db_host_var: Optional[str] = self.args.db_host if self.args else None
        if db_host_var is None:
            # read from secrets_file
            logger.debug(f"Reading DATABASE_HOST from secrets")
            secret_data = self.get_secret_data()
            if secret_data is not None:
                db_host_var = secret_data.get("DATABASE_HOST", db_host_var)
        return db_host_var

    def get_db_port(self) -> Optional[str]:
        db_port_var: Optional[Union[int, str]] = (
            self.args.db_port if self.args else None
        )
        if db_port_var is None:
            # read from secrets_file
            logger.debug(f"Reading DATABASE_PORT from secrets")
            secret_data = self.get_secret_data()
            if secret_data is not None:
                db_port_var = secret_data.get("DATABASE_PORT", db_port_var)
        return str(db_port_var) if db_port_var is not None else db_port_var

    def get_db_dialect(self) -> Optional[str]:
        db_dialect_var: Optional[str] = self.args.db_dialect if self.args else None
        if db_dialect_var is None:
            # read from secrets_file
            logger.debug(f"Reading DATABASE_DIALECT from secrets")
            secret_data = self.get_secret_data()
            if secret_data is not None:
                db_dialect_var = secret_data.get("DATABASE_DIALECT", db_dialect_var)
        return db_dialect_var

    def get_redis_host(self) -> Optional[str]:
        redis_host_var: Optional[str] = self.args.redis_host if self.args else None
        if redis_host_var is None:
            # read from secrets_file
            logger.debug(f"Reading REDIS_HOST from secrets")
            secret_data = self.get_secret_data()
            if secret_data is not None:
                redis_host_var = secret_data.get("REDIS_HOST", redis_host_var)
        return redis_host_var

    def get_redis_port(self) -> Optional[str]:
        redis_port_var: Optional[Union[int, str]] = (
            self.args.redis_port if self.args else None
        )
        if redis_port_var is None:
            # read from secrets_file
            logger.debug(f"Reading REDIS_PORT from secrets")
            secret_data = self.get_secret_data()
            if secret_data is not None:
                redis_port_var = secret_data.get("REDIS_PORT", redis_port_var)
        return str(redis_port_var) if redis_port_var is not None else redis_port_var

    def get_redis_driver(self) -> Optional[str]:
        redis_driver_var: Optional[str] = self.args.redis_driver if self.args else None
        if redis_driver_var is None:
            # read from secrets_file
            logger.debug(f"Reading REDIS_DRIVER from secrets")
            secret_data = self.get_secret_data()
            if secret_data is not None:
                redis_driver_var = secret_data.get("REDIS_DRIVER", redis_driver_var)
        return redis_driver_var

    ######################################################
    ## Docker Resources
    ######################################################

    def get_docker_rg(
        self, docker_build_context: DockerBuildContext
    ) -> Optional[DockerResourceGroup]:

        app_name = self.args.name
        logger.debug(f"Building {app_name} DockerResourceGroup")

        # Workspace paths
        if self.workspace_root_path is None:
            logger.error("Invalid workspace_root_path")
            return None
        workspace_name = self.workspace_root_path.stem
        workspace_root_container_path = Path(
            self.args.workspace_mount_container_path
        ).joinpath(workspace_name)
        requirements_file_container_path = workspace_root_container_path.joinpath(
            self.args.requirements_file_path
        )
        scripts_dir_container_path = (
            workspace_root_container_path.joinpath(self.scripts_dir)
            if self.scripts_dir
            else None
        )
        storage_dir_container_path = (
            workspace_root_container_path.joinpath(self.storage_dir)
            if self.storage_dir
            else None
        )
        meta_dir_container_path = (
            workspace_root_container_path.joinpath(self.meta_dir)
            if self.meta_dir
            else None
        )
        products_dir_container_path = (
            workspace_root_container_path.joinpath(self.products_dir)
            if self.products_dir
            else None
        )
        notebooks_dir_container_path = (
            workspace_root_container_path.joinpath(self.notebooks_dir)
            if self.notebooks_dir
            else None
        )
        workspace_config_dir_container_path = (
            workspace_root_container_path.joinpath(self.workspace_config_dir)
            if self.workspace_config_dir
            else None
        )

        # Container pythonpath
        python_path = (
            self.args.python_path
            or f"/app/pythonpath:{self.args.resources_dir_container_path}/pythonpath_dev"
        )

        # Container Environment
        container_env: Dict[str, str] = {
            # Env variables used by data workflows and data assets
            "PHI_WORKSPACE_MOUNT": str(self.args.workspace_mount_container_path),
            "PHI_WORKSPACE_ROOT": str(workspace_root_container_path),
            "PYTHONPATH": python_path,
            PHIDATA_RUNTIME_ENV_VAR: "docker",
            SCRIPTS_DIR_ENV_VAR: str(scripts_dir_container_path),
            STORAGE_DIR_ENV_VAR: str(storage_dir_container_path),
            META_DIR_ENV_VAR: str(meta_dir_container_path),
            PRODUCTS_DIR_ENV_VAR: str(products_dir_container_path),
            NOTEBOOKS_DIR_ENV_VAR: str(notebooks_dir_container_path),
            WORKSPACE_CONFIG_DIR_ENV_VAR: str(workspace_config_dir_container_path),
            "INSTALL_REQUIREMENTS": str(self.args.install_requirements),
            "REQUIREMENTS_FILE_PATH": str(requirements_file_container_path),
            "REQUIREMENTS_LOCAL": str(requirements_file_container_path),
            "MOUNT_WORKSPACE": str(self.args.mount_workspace),
            "MOUNT_RESOURCES": str(self.args.mount_resources),
            "WAIT_FOR_DB": str(self.args.wait_for_db),
            "WAIT_FOR_REDIS": str(self.args.wait_for_redis),
            "SUPERSET_LOAD_EXAMPLES": "yes" if self.args.load_examples else "no",
            # Print env when the container starts
            "PRINT_ENV_ON_LOAD": str(self.args.print_env_on_load),
        }

        # Set airflow env vars
        self.set_aws_env_vars(env_dict=container_env)

        if self.args.superset_config_path is not None:
            container_env["SUPERSET_CONFIG_PATH"] = self.args.superset_config_path

        if self.args.flask_env is not None:
            container_env["FLASK_ENV"] = self.args.flask_env

        if self.args.superset_env is not None:
            container_env["SUPERSET_ENV"] = self.args.superset_env

        # Superset db connection
        db_user = self.get_db_user()
        db_password = self.get_db_password()
        db_schema = self.get_db_schema()
        db_host = self.get_db_host()
        db_port = self.get_db_port()
        db_dialect = self.get_db_dialect()
        if self.args.db_app is not None and isinstance(self.args.db_app, DbApp):
            logger.debug(f"Reading db connection details from: {self.args.db_app.name}")
            if db_user is None:
                db_user = self.args.db_app.get_db_user()
            if db_password is None:
                db_password = self.args.db_app.get_db_password()
            if db_schema is None:
                db_schema = self.args.db_app.get_db_schema()
            if db_host is None:
                db_host = self.args.db_app.get_db_host_docker()
            if db_port is None:
                db_port = str(self.args.db_app.get_db_port_docker())
            if db_dialect is None:
                db_dialect = self.args.db_app.get_db_driver()

        if db_user is not None:
            container_env["DATABASE_USER"] = db_user
        # Ideally we dont want the password in the env
        # But the superest image expects it :(
        if db_password is not None:
            container_env["DATABASE_PASSWORD"] = db_password
        if db_schema is not None:
            container_env["DATABASE_DB"] = db_schema
        if db_host is not None:
            container_env["DATABASE_HOST"] = db_host
        if db_port is not None:
            container_env["DATABASE_PORT"] = str(db_port)
        if db_dialect is not None:
            container_env["DATABASE_DIALECT"] = db_dialect

        # Superset redis connection
        redis_host = self.get_redis_host()
        redis_port = self.get_redis_port()
        redis_driver = self.get_redis_driver()
        if self.args.redis_app is not None and isinstance(self.args.redis_app, DbApp):
            logger.debug(
                f"Reading redis connection details from: {self.args.redis_app.name}"
            )
            if redis_host is None:
                redis_host = self.args.redis_app.get_db_host_docker()
            if redis_port is None:
                redis_port = str(self.args.redis_app.get_db_port_docker())
            if redis_driver is None:
                redis_driver = self.args.redis_app.get_db_driver()

        if redis_host is not None:
            container_env["REDIS_HOST"] = redis_host
        if redis_port is not None:
            container_env["REDIS_PORT"] = str(redis_port)
        if redis_driver is not None:
            container_env["REDIS_DRIVER"] = str(redis_driver)

        # Update the container env using env_file
        env_data_from_file = self.get_env_data()
        if env_data_from_file is not None:
            container_env.update(env_data_from_file)

        # Update the container env using secrets_file or a secrets backend
        secret_data_from_file = self.get_secret_data()
        if secret_data_from_file is not None:
            container_env.update(secret_data_from_file)

        # Update the container env with user provided env, this overwrites any existing variables
        if self.args.env is not None and isinstance(self.args.env, dict):
            container_env.update(self.args.env)

        # Container Volumes
        # container_volumes is a dictionary which configures the volumes to mount
        # inside the container. The key is either the host path or a volume name,
        # and the value is a dictionary with 2 keys:
        #   bind - The path to mount the volume inside the container
        #   mode - Either rw to mount the volume read/write, or ro to mount it read-only.
        # For example:
        # {
        #   '/home/user1/': {'bind': '/mnt/vol2', 'mode': 'rw'},
        #   '/var/www': {'bind': '/mnt/vol1', 'mode': 'ro'}
        # }
        container_volumes = self.args.container_volumes_docker or {}
        # Create a volume for the workspace dir
        if self.args.mount_workspace:
            workspace_root_path_str = str(self.workspace_root_path)
            workspace_root_container_path_str = str(workspace_root_container_path)
            logger.debug(f"Mounting: {workspace_root_path_str}")
            logger.debug(f"\tto: {workspace_root_container_path_str}")
            container_volumes[workspace_root_path_str] = {
                "bind": workspace_root_container_path_str,
                "mode": "rw",
            }
        # Create a volume for the resources
        if self.args.mount_resources:
            resources_dir_path = str(
                self.workspace_root_path.joinpath(self.args.resources_dir)
            )
            logger.debug(f"Mounting: {resources_dir_path}")
            logger.debug(f"\tto: {self.args.resources_dir_container_path}")
            container_volumes[resources_dir_path] = {
                "bind": self.args.resources_dir_container_path,
                "mode": "ro",
            }

        # Container Ports
        # container_ports is a dictionary which configures the ports to bind
        # inside the container. The key is the port to bind inside the container
        #   either as an integer or a string in the form port/protocol
        # and the value is the corresponding port to open on the host.
        # For example:
        #   {'2222/tcp': 3333} will expose port 2222 inside the container as port 3333 on the host.
        container_ports: Dict[str, int] = self.args.container_ports_docker or {}

        # if open_container_port = True
        if self.args.open_container_port:
            container_ports[
                str(self.args.container_port)
            ] = self.args.container_host_port

        # if open_app_port = True
        # 1. Set the app_port in the container env
        # 2. Open the app_port
        if self.args.open_app_port:
            # Set the app port in the container_env
            container_env["SUPERSET_PORT"] = str(self.args.app_port)
            # Open the port
            container_ports[str(self.args.app_port)] = self.args.app_host_port

        # Create the container
        docker_container = DockerContainer(
            name=self.get_container_name(),
            image=get_image_str(self.args.image_name, self.args.image_tag),
            entrypoint=self.args.entrypoint,
            command=self.args.command,
            detach=self.args.container_detach,
            auto_remove=self.args.container_auto_remove,
            healthcheck=self.args.container_healthcheck,
            hostname=self.args.container_hostname,
            labels=self.args.container_labels,
            environment=container_env,
            network=docker_build_context.network,
            platform=self.args.container_platform,
            ports=container_ports if len(container_ports) > 0 else None,
            remove=self.args.container_remove,
            restart_policy=self.args.container_restart_policy_docker,
            stdin_open=self.args.container_stdin_open,
            tty=self.args.container_tty,
            user=self.args.container_user,
            volumes=container_volumes,
            working_dir=self.args.container_working_dir,
            use_cache=self.args.use_cache,
        )
        # logger.debug(f"Container Env: {docker_container.environment}")

        docker_rg = DockerResourceGroup(
            name=app_name,
            enabled=self.args.enabled,
            network=DockerNetwork(name=docker_build_context.network),
            containers=[docker_container],
        )
        return docker_rg

    def init_docker_resource_groups(
        self, docker_build_context: DockerBuildContext
    ) -> None:
        docker_rg = self.get_docker_rg(docker_build_context)
        if docker_rg is not None:
            if self.docker_resource_groups is None:
                self.docker_resource_groups = OrderedDict()
            self.docker_resource_groups[docker_rg.name] = docker_rg

    ######################################################
    ## K8s Resources
    ######################################################

    def get_k8s_rg(
        self, k8s_build_context: K8sBuildContext
    ) -> Optional[K8sResourceGroup]:

        app_name = self.args.name
        logger.debug(f"Building {app_name} K8sResourceGroup")

        # Workspace paths
        if self.workspace_root_path is None:
            logger.error("Invalid workspace_root_path")
            return None
        workspace_name = self.workspace_root_path.stem
        workspace_root_container_path = Path(
            self.args.workspace_mount_container_path
        ).joinpath(workspace_name)
        requirements_file_container_path = workspace_root_container_path.joinpath(
            self.args.requirements_file_path
        )
        scripts_dir_container_path = (
            workspace_root_container_path.joinpath(self.scripts_dir)
            if self.scripts_dir
            else None
        )
        storage_dir_container_path = (
            workspace_root_container_path.joinpath(self.storage_dir)
            if self.storage_dir
            else None
        )
        meta_dir_container_path = (
            workspace_root_container_path.joinpath(self.meta_dir)
            if self.meta_dir
            else None
        )
        products_dir_container_path = (
            workspace_root_container_path.joinpath(self.products_dir)
            if self.products_dir
            else None
        )
        notebooks_dir_container_path = (
            workspace_root_container_path.joinpath(self.notebooks_dir)
            if self.notebooks_dir
            else None
        )
        workspace_config_dir_container_path = (
            workspace_root_container_path.joinpath(self.workspace_config_dir)
            if self.workspace_config_dir
            else None
        )

        # Init K8s resources for the CreateK8sResourceGroup
        ns: Optional[CreateNamespace] = self.args.namespace
        sa: Optional[CreateServiceAccount] = self.args.service_account
        cr: Optional[CreateClusterRole] = self.args.cluster_role
        crb: Optional[CreateClusterRoleBinding] = self.args.cluster_role_binding
        secrets: List[CreateSecret] = self.args.extra_secrets or []
        config_maps: List[CreateConfigMap] = self.args.extra_config_maps or []
        storage_classes: List[CreateStorageClass] = (
            self.args.extra_storage_classes or []
        )
        services: List[CreateService] = self.args.extra_services or []
        deployments: List[CreateDeployment] = self.args.extra_deployments or []
        custom_objects: List[CreateCustomObject] = self.args.extra_custom_objects or []
        crds: List[CreateCustomResourceDefinition] = self.args.extra_crds or []
        pvs: List[CreatePersistentVolume] = self.args.extra_pvs or []
        pvcs: List[CreatePVC] = self.args.extra_pvcs or []
        containers: List[CreateContainer] = self.args.extra_containers or []
        init_containers: List[CreateContainer] = self.args.extra_init_containers or []
        ports: List[CreatePort] = self.args.extra_ports or []
        volumes: List[CreateVolume] = self.args.extra_volumes or []

        # Common variables used by all resources
        ns_name: str = self.args.ns_name or k8s_build_context.namespace
        sa_name: Optional[str] = (
            self.args.sa_name or k8s_build_context.service_account_name
        )
        common_labels: Optional[Dict[str, str]] = k8s_build_context.labels

        # -*- Define RBAC resources
        # WebUI/Scheduler pods should run with serviceAccount which have RBAC
        # permissions on the k8s cluster to get logs
        # https://github.com/apache/airflow/issues/11696#issuecomment-715886117
        if self.args.use_rbac:
            if ns is None:
                ns = CreateNamespace(
                    ns=ns_name,
                    app_name=app_name,
                    labels=common_labels,
                )
            ns_name = ns.ns
            if sa is None:
                sa = CreateServiceAccount(
                    sa_name=sa_name or get_default_sa_name(app_name),
                    app_name=app_name,
                    namespace=ns_name,
                )
            sa_name = sa.sa_name
            if cr is None:
                cr = CreateClusterRole(
                    cr_name=self.args.cr_name or get_default_cr_name(app_name),
                    rules=[
                        PolicyRule(
                            api_groups=[""],
                            resources=[
                                "pods",
                                "secrets",
                                "configmaps",
                            ],
                            verbs=[
                                "get",
                                "list",
                                "watch",
                                "create",
                                "update",
                                "patch",
                                "delete",
                            ],
                        ),
                        PolicyRule(
                            api_groups=[""],
                            resources=[
                                "pods/logs",
                            ],
                            verbs=[
                                "get",
                                "list",
                            ],
                        ),
                        # PolicyRule(
                        #     api_groups=[""],
                        #     resources=[
                        #         "pods/exec",
                        #     ],
                        #     verbs=[
                        #         "get",
                        #         "create",
                        #     ],
                        # ),
                    ],
                    app_name=app_name,
                    labels=common_labels,
                )
            if crb is None:
                crb = CreateClusterRoleBinding(
                    crb_name=self.args.crb_name or get_default_crb_name(app_name),
                    cr_name=cr.cr_name,
                    service_account_name=sa.sa_name,
                    app_name=app_name,
                    namespace=ns_name,
                    labels=common_labels,
                )

        # Container pythonpath
        python_path = (
            self.args.python_path
            or f"/app/pythonpath:{self.args.resources_dir_container_path}/pythonpath_dev"
        )

        # Container Environment
        container_env: Dict[str, str] = {
            # Env variables used by data workflows and data assets
            "PHI_WORKSPACE_MOUNT": str(self.args.workspace_mount_container_path),
            "PHI_WORKSPACE_ROOT": str(workspace_root_container_path),
            "PYTHONPATH": python_path,
            PHIDATA_RUNTIME_ENV_VAR: "kubernetes",
            SCRIPTS_DIR_ENV_VAR: str(scripts_dir_container_path),
            STORAGE_DIR_ENV_VAR: str(storage_dir_container_path),
            META_DIR_ENV_VAR: str(meta_dir_container_path),
            PRODUCTS_DIR_ENV_VAR: str(products_dir_container_path),
            NOTEBOOKS_DIR_ENV_VAR: str(notebooks_dir_container_path),
            WORKSPACE_CONFIG_DIR_ENV_VAR: str(workspace_config_dir_container_path),
            "INSTALL_REQUIREMENTS": str(self.args.install_requirements),
            "REQUIREMENTS_FILE_PATH": str(requirements_file_container_path),
            "REQUIREMENTS_LOCAL": str(requirements_file_container_path),
            "MOUNT_WORKSPACE": str(self.args.mount_workspace),
            "MOUNT_RESOURCES": str(self.args.mount_resources),
            "WAIT_FOR_DB": str(self.args.wait_for_db),
            "WAIT_FOR_REDIS": str(self.args.wait_for_redis),
            "SUPERSET_LOAD_EXAMPLES": "yes" if self.args.load_examples else "no",
            # Print env when the container starts
            "PRINT_ENV_ON_LOAD": str(self.args.print_env_on_load),
        }

        # Set airflow env vars
        self.set_aws_env_vars(env_dict=container_env)

        if self.args.superset_config_path is not None:
            container_env["SUPERSET_CONFIG_PATH"] = self.args.superset_config_path

        if self.args.flask_env is not None:
            container_env["FLASK_ENV"] = self.args.flask_env

        if self.args.superset_env is not None:
            container_env["SUPERSET_ENV"] = self.args.superset_env

        # Superset db connection
        db_user = self.get_db_user()
        db_password = self.get_db_password()
        db_schema = self.get_db_schema()
        db_host = self.get_db_host()
        db_port = self.get_db_port()
        db_dialect = self.get_db_dialect()
        if self.args.db_app is not None and isinstance(self.args.db_app, DbApp):
            logger.debug(f"Reading db connection details from: {self.args.db_app.name}")
            if db_user is None:
                db_user = self.args.db_app.get_db_user()
            if db_password is None:
                db_password = self.args.db_app.get_db_password()
            if db_schema is None:
                db_schema = self.args.db_app.get_db_schema()
            if db_host is None:
                db_host = self.args.db_app.get_db_host_k8s()
            if db_port is None:
                db_port = str(self.args.db_app.get_db_port_k8s())
            if db_dialect is None:
                db_dialect = self.args.db_app.get_db_driver()

        if db_user is not None:
            container_env["DATABASE_USER"] = db_user
        # Ideally we dont want the password in the env
        # But the superest image expects it :(
        if db_password is not None:
            container_env["DATABASE_PASSWORD"] = db_password
        if db_schema is not None:
            container_env["DATABASE_DB"] = db_schema
        if db_host is not None:
            container_env["DATABASE_HOST"] = db_host
        if db_port is not None:
            container_env["DATABASE_PORT"] = str(db_port)
        if db_dialect is not None:
            container_env["DATABASE_DIALECT"] = db_dialect

        # Superset redis connection
        redis_host = self.get_redis_host()
        redis_port = self.get_redis_port()
        redis_driver = self.get_redis_driver()
        if self.args.redis_app is not None and isinstance(self.args.redis_app, DbApp):
            logger.debug(
                f"Reading redis connection details from: {self.args.redis_app.name}"
            )
            if redis_host is None:
                redis_host = self.args.redis_app.get_db_host_k8s()
            if redis_port is None:
                redis_port = str(self.args.redis_app.get_db_port_k8s())
            if redis_driver is None:
                redis_driver = self.args.redis_app.get_db_driver()

        if redis_host is not None:
            container_env["REDIS_HOST"] = redis_host
        if redis_port is not None:
            container_env["REDIS_PORT"] = str(redis_port)
        if redis_driver is not None:
            container_env["REDIS_DRIVER"] = str(redis_driver)

        # Update the container env using env_file
        env_data_from_file = self.get_env_data()
        if env_data_from_file is not None:
            container_env.update(env_data_from_file)

        # Update the container env with user provided env, this overwrites any existing variables
        if self.args.env is not None and isinstance(self.args.env, dict):
            container_env.update(self.args.env)

        # Create a ConfigMap to set the container env variables which are not Secret
        container_env_cm = CreateConfigMap(
            cm_name=self.args.config_map_name or get_default_configmap_name(app_name),
            app_name=app_name,
            data=container_env,
        )
        # logger.debug(f"ConfigMap {container_env_cm.cm_name}: {container_env_cm.json(indent=2)}")
        config_maps.append(container_env_cm)

        # Create a Secret to set the container env variables which are Secret
        _secret_data = self.get_secret_data()
        if _secret_data is not None:
            container_env_secret = CreateSecret(
                secret_name=self.args.secret_name or get_default_secret_name(app_name),
                app_name=app_name,
                string_data=_secret_data,
            )
            secrets.append(container_env_secret)

        # Container Volumes
        # If mount_workspace=True first check if the workspace
        # should be mounted locally, otherwise
        # Create a Sidecar git-sync container and volume
        if self.args.mount_workspace:
            workspace_volume_name = (
                self.args.workspace_volume_name or get_default_volume_name(app_name)
            )

            if self.args.k8s_mount_local_workspace:
                workspace_root_path_str = str(self.workspace_root_path)
                workspace_root_container_path_str = str(workspace_root_container_path)
                logger.debug(f"Mounting: {workspace_root_path_str}")
                logger.debug(f"\tto: {workspace_root_container_path_str}")
                workspace_volume = CreateVolume(
                    volume_name=workspace_volume_name,
                    app_name=app_name,
                    mount_path=workspace_root_container_path_str,
                    volume_type=VolumeType.HOST_PATH,
                    host_path=HostPathVolumeSource(
                        path=workspace_root_path_str,
                    ),
                )
                volumes.append(workspace_volume)

            elif self.args.create_git_sync_sidecar:
                workspace_mount_container_path_str = str(
                    self.args.workspace_mount_container_path
                )
                logger.debug(f"Creating EmptyDir")
                logger.debug(f"\tat: {workspace_mount_container_path_str}")
                workspace_volume = CreateVolume(
                    volume_name=workspace_volume_name,
                    app_name=app_name,
                    mount_path=workspace_mount_container_path_str,
                    volume_type=VolumeType.EMPTY_DIR,
                )
                volumes.append(workspace_volume)

                if self.args.git_sync_repo is not None:
                    git_sync_env = {
                        "GIT_SYNC_REPO": self.args.git_sync_repo,
                        "GIT_SYNC_ROOT": str(self.args.workspace_mount_container_path),
                        "GIT_SYNC_DEST": workspace_name,
                    }
                    if self.args.git_sync_branch is not None:
                        git_sync_env["GIT_SYNC_BRANCH"] = self.args.git_sync_branch
                    if self.args.git_sync_wait is not None:
                        git_sync_env["GIT_SYNC_WAIT"] = str(self.args.git_sync_wait)
                    git_sync_container = CreateContainer(
                        container_name="git-sync-workspaces",
                        app_name=app_name,
                        image_name="k8s.gcr.io/git-sync",
                        image_tag="v3.1.1",
                        env=git_sync_env,
                        envs_from_configmap=[cm.cm_name for cm in config_maps]
                        if len(config_maps) > 0
                        else None,
                        envs_from_secret=[secret.secret_name for secret in secrets]
                        if len(secrets) > 0
                        else None,
                        volumes=[workspace_volume],
                    )
                    containers.append(git_sync_container)

                    if self.args.create_git_sync_init_container:
                        git_sync_init_env: Dict[str, Any] = {"GIT_SYNC_ONE_TIME": True}
                        git_sync_init_env.update(git_sync_env)
                        _git_sync_init_container = CreateContainer(
                            container_name="git-sync-init",
                            app_name=git_sync_container.app_name,
                            image_name=git_sync_container.image_name,
                            image_tag=git_sync_container.image_tag,
                            env=git_sync_init_env,
                            envs_from_configmap=git_sync_container.envs_from_configmap,
                            envs_from_secret=git_sync_container.envs_from_secret,
                            volumes=git_sync_container.volumes,
                        )
                        init_containers.append(_git_sync_init_container)
                else:
                    logger.error("GIT_SYNC_REPO invalid")

        # Create the ports to open
        # if open_container_port = True
        if self.args.open_container_port:
            container_port = CreatePort(
                name=self.args.container_port_name,
                container_port=self.args.container_port,
            )
            ports.append(container_port)

        # if open_app_port = True
        # 1. Set the app_port in the container env
        # 2. Open the superset app port
        app_port: Optional[CreatePort] = None
        if self.args.open_app_port:
            # Set the app port in the container env
            if container_env_cm.data is not None:
                container_env_cm.data["SUPERSET_PORT"] = str(self.args.app_port)
            # Open the port
            app_port = CreatePort(
                name=self.args.app_port_name,
                container_port=self.args.app_port,
                service_port=self.get_app_service_port(),
                node_port=self.args.app_node_port,
                target_port=self.args.app_target_port or self.args.app_port_name,
            )
            ports.append(app_port)

        container_labels: Dict[str, Any] = common_labels or {}
        if self.args.container_labels is not None and isinstance(
            self.args.container_labels, dict
        ):
            container_labels.update(self.args.container_labels)

        # Create the container
        superset_container = CreateContainer(
            container_name=self.get_container_name(),
            app_name=app_name,
            image_name=self.args.image_name,
            image_tag=self.args.image_tag,
            # Equivalent to docker images CMD
            args=[self.args.command]
            if isinstance(self.args.command, str)
            else self.args.command,
            # Equivalent to docker images ENTRYPOINT
            command=[self.args.entrypoint]
            if isinstance(self.args.entrypoint, str)
            else self.args.entrypoint,
            image_pull_policy=self.args.image_pull_policy,
            envs_from_configmap=[cm.cm_name for cm in config_maps]
            if len(config_maps) > 0
            else None,
            envs_from_secret=[secret.secret_name for secret in secrets]
            if len(secrets) > 0
            else None,
            ports=ports if len(ports) > 0 else None,
            volumes=volumes if len(volumes) > 0 else None,
            labels=container_labels,
        )
        containers.append(superset_container)

        # Set default container for kubectl commands
        # https://kubernetes.io/docs/reference/labels-annotations-taints/#kubectl-kubernetes-io-default-container
        pod_annotations = {
            "kubectl.kubernetes.io/default-container": superset_container.container_name
        }
        if self.args.pod_annotations is not None and isinstance(
            self.args.pod_annotations, dict
        ):
            pod_annotations.update(self.args.pod_annotations)

        deploy_labels: Dict[str, Any] = common_labels or {}
        if self.args.deploy_labels is not None and isinstance(
            self.args.deploy_labels, dict
        ):
            deploy_labels.update(self.args.deploy_labels)

        # Create the deployment
        superset_deployment = CreateDeployment(
            deploy_name=self.args.deploy_name or get_default_deploy_name(app_name),
            pod_name=self.args.pod_name or get_default_pod_name(app_name),
            app_name=app_name,
            namespace=ns_name,
            service_account_name=sa_name,
            replicas=self.args.replicas,
            containers=containers,
            init_containers=init_containers if len(init_containers) > 0 else None,
            pod_node_selector=self.args.pod_node_selector,
            restart_policy=self.args.deploy_restart_policy,
            termination_grace_period_seconds=self.args.termination_grace_period_seconds,
            volumes=volumes if len(volumes) > 0 else None,
            labels=deploy_labels,
            pod_annotations=pod_annotations,
            topology_spread_key=self.args.topology_spread_key,
            topology_spread_max_skew=self.args.topology_spread_max_skew,
            topology_spread_when_unsatisfiable=self.args.topology_spread_when_unsatisfiable,
        )
        deployments.append(superset_deployment)

        # Create the services
        if self.args.create_app_service:
            app_svc_labels: Dict[str, Any] = common_labels or {}
            if self.args.app_svc_labels is not None and isinstance(
                self.args.app_svc_labels, dict
            ):
                app_svc_labels.update(self.args.app_svc_labels)

            app_service = CreateService(
                service_name=self.get_app_service_name(),
                app_name=app_name,
                namespace=ns_name,
                service_account_name=sa_name,
                service_type=self.args.app_svc_type,
                deployment=superset_deployment,
                ports=[app_port] if app_port else None,
                labels=app_svc_labels,
            )
            services.append(app_service)

        # Create the K8sResourceGroup
        k8s_resource_group = CreateK8sResourceGroup(
            name=app_name,
            enabled=self.args.enabled,
            ns=ns,
            sa=sa,
            cr=cr,
            crb=crb,
            secrets=secrets if len(secrets) > 0 else None,
            config_maps=config_maps if len(config_maps) > 0 else None,
            storage_classes=storage_classes if len(storage_classes) > 0 else None,
            services=services if len(services) > 0 else None,
            deployments=deployments if len(deployments) > 0 else None,
            custom_objects=custom_objects if len(custom_objects) > 0 else None,
            crds=crds if len(crds) > 0 else None,
            pvs=pvs if len(pvs) > 0 else None,
            pvcs=pvcs if len(pvcs) > 0 else None,
        )

        return k8s_resource_group.create()

    def init_k8s_resource_groups(self, k8s_build_context: K8sBuildContext) -> None:
        k8s_rg = self.get_k8s_rg(k8s_build_context)
        if k8s_rg is not None:
            if self.k8s_resource_groups is None:
                self.k8s_resource_groups = OrderedDict()
            self.k8s_resource_groups[k8s_rg.name] = k8s_rg



from phidata.app.airflow.airflow_webserver import (
    AirflowWebserver,
    ImagePullPolicy,
    ServiceType,
)
from phidata.app.airflow.airflow_scheduler import AirflowScheduler
from phidata.app.airflow.airflow_worker import AirflowWorker
from phidata.app.airflow.airflow_flower import AirflowFlower
from phidata.app.airflow.airflow_manager import AirflowManager



from phidata.infra.base import InfraConfig
from phidata.workspace import WorkspaceConfig

from phidata.utils.log import logger


def prep_infra_config(
    infra_config: InfraConfig,
    ws_config: WorkspaceConfig,
) -> InfraConfig:

    logger.debug(f"Updating {infra_config.__class__.__name__} using WorkspaceConfig")

    ######################################################################
    # NOTE: VERY IMPORTANT TO GET RIGHT
    # Prep InfraConfig using the WorkspaceConfig
    # 1. Pass down the paths from the WorkspaceConfig
    #       These paths are used everywhere from Infra, Apps, Resources, Workflows
    # 2. Pass down local_env, docker_env, k8s_env
    # 3. Pass down common cloud configuration. eg: aws_region, aws_profile
    ######################################################################

    # -*- Path parameters
    infra_config.scripts_dir = ws_config.scripts_dir
    infra_config.storage_dir = ws_config.storage_dir
    infra_config.meta_dir = ws_config.meta_dir
    infra_config.products_dir = ws_config.products_dir
    infra_config.notebooks_dir = ws_config.notebooks_dir
    # The workspace_root_path is the ROOT directory for the workspace
    infra_config.workspace_root_path = ws_config.workspace_root_path
    infra_config.workspace_config_dir = ws_config.workspace_config_dir
    infra_config.workspace_config_file_path = ws_config.workspace_config_file_path

    # -*- Environment parameters
    # only update the param if they are not available.
    # i.e. prefer the configs param if provided
    if infra_config.local_env is None and ws_config.local_env is not None:
        infra_config.local_env = ws_config.local_env
    if infra_config.local_env_file is None and ws_config.local_env_file is not None:
        infra_config.local_env_file = ws_config.local_env_file
    if infra_config.docker_env is None and ws_config.docker_env is not None:
        infra_config.docker_env = ws_config.docker_env
    if infra_config.docker_env_file is None and ws_config.docker_env_file is not None:
        infra_config.docker_env_file = ws_config.docker_env_file
    if infra_config.k8s_env is None and ws_config.k8s_env is not None:
        infra_config.k8s_env = ws_config.k8s_env
    if infra_config.k8s_env_file is None and ws_config.k8s_env_file is not None:
        infra_config.k8s_env_file = ws_config.k8s_env_file

    # -*- AWS parameters
    # only update the param if they are available.
    # i.e. prefer the configs param if provided
    if infra_config.aws_region is None and ws_config.aws_region is not None:
        infra_config.aws_region = ws_config.aws_region
    if infra_config.aws_profile is None and ws_config.aws_profile is not None:
        infra_config.aws_profile = ws_config.aws_profile
    if infra_config.aws_config_file is None and ws_config.aws_config_file is not None:
        infra_config.aws_config_file = ws_config.aws_config_file
    if (
        infra_config.aws_shared_credentials_file is None
        and ws_config.aws_shared_credentials_file is not None
    ):
        infra_config.aws_shared_credentials_file = ws_config.aws_shared_credentials_file

    # -*- `phi` cli parameters
    # only update the param if they are available.
    # i.e. prefer the configs param if provided
    if (
        infra_config.continue_on_create_failure is None
        and ws_config.continue_on_create_failure is not None
    ):
        infra_config.continue_on_create_failure = ws_config.continue_on_create_failure
    if (
        infra_config.continue_on_delete_failure is None
        and ws_config.continue_on_delete_failure is not None
    ):
        infra_config.continue_on_delete_failure = ws_config.continue_on_delete_failure
    if (
        infra_config.continue_on_patch_failure is None
        and ws_config.continue_on_patch_failure is not None
    ):
        infra_config.continue_on_patch_failure = ws_config.continue_on_patch_failure

    return infra_config



from phidata.infra.k8s.enums.api_group import *
from phidata.infra.k8s.enums.api_version import *
from phidata.infra.k8s.enums.image_pull_policy import *
from phidata.infra.k8s.enums.kind import *
from phidata.infra.k8s.enums.manager_status import *
from phidata.infra.k8s.enums.protocol import *
from phidata.infra.k8s.enums.pv import *
from phidata.infra.k8s.enums.restart_policy import *
from phidata.infra.k8s.enums.service_type import *
from phidata.infra.k8s.enums.storage_class import *
from phidata.infra.k8s.enums.volume_type import *



import yaml
from collections import OrderedDict
from pathlib import Path
from typing import Optional, Dict, List, Any

from phidata.app import PhidataApp, PhidataAppArgs
from phidata.infra.docker.resource.group import DockerBuildContext
from phidata.infra.k8s.resource.group import (
    K8sResourceGroup,
    K8sBuildContext,
)
from phidata.utils.log import logger


class K8sAppArgs(PhidataAppArgs):
    name: str = "K8sApp"
    version: str = "1"
    enabled: bool = True

    manifest_dir: Optional[Path] = None
    manifest_file: Optional[Path] = None
    cache_manifests: bool = True


class K8sApp(PhidataApp):
    def __init__(
        self,
        name: str = "K8sApp",
        version: str = "1",
        enabled: bool = True,
        manifest_dir: Optional[Path] = None,
        manifest_file: Optional[Path] = None,
        cache_manifests: bool = True,
    ):
        super().__init__()
        try:
            self.args: K8sAppArgs = K8sAppArgs(
                name=name,
                version=version,
                enabled=enabled,
                manifest_dir=manifest_dir,
                manifest_file=manifest_file,
                cache_manifests=cache_manifests,
            )
        except Exception as e:
            logger.error(f"Args for {self.__class__.__name__} are not valid")
            raise

    def init_docker_resource_groups(
        self, docker_build_context: DockerBuildContext
    ) -> None:
        logger.error(f"{self.args.name} not available for Docker")

    ######################################################
    ## K8s Resources
    ######################################################

    def get_k8s_rg(
        self, k8s_build_context: K8sBuildContext
    ) -> Optional[K8sResourceGroup]:
        app_name = self.name
        logger.debug(f"Building {app_name} K8sResourceGroup")

        # K8sResourceGroup to return
        k8s_resource_group: K8sResourceGroup = K8sResourceGroup(
            name=app_name,
            enabled=self.enabled,
        )

        manifest_dir: Optional[Path] = self.args.manifest_dir
        manifest_file: Optional[Path] = self.args.manifest_file
        manifests_to_parse: List[Dict[str, Any]] = []

        # Read manifests from file
        if manifest_file is not None:
            logger.debug(f"Reading manifests from {manifest_file}")
            suffix = manifest_file.suffix
            if suffix == ".yaml":
                for manifest in yaml.safe_load_all(manifest_file.read_text()):
                    if isinstance(manifest, dict):
                        manifests_to_parse.append(manifest)
            else:
                logger.error(f"{suffix} files not yet supported")

            if len(manifests_to_parse) == 0:
                logger.warning(f"No manifests found in {str(manifest_file)}")
                return None

        # Read manifests from dir
        elif manifest_dir is not None:
            if manifest_dir.exists() and manifest_dir.is_dir():
                for _file in manifest_dir.glob("*.yaml"):
                    logger.debug(f"Reading manifests from {_file}")
                    for manifest in yaml.safe_load_all(_file.read_text()):
                        if isinstance(manifest, dict):
                            manifests_to_parse.append(manifest)

            if len(manifests_to_parse) == 0:
                logger.warning(f"No manifests found in {str(manifest_dir)}")
                return None

        for manifest in manifests_to_parse:
            k8s_resource_group.add_manifest_to_group(manifest)

        return k8s_resource_group

    def init_k8s_resource_groups(self, k8s_build_context: K8sBuildContext) -> None:
        k8s_rg = self.get_k8s_rg(k8s_build_context)
        if k8s_rg is not None:
            if self.k8s_resource_groups is None:
                self.k8s_resource_groups = OrderedDict()
            self.k8s_resource_groups[k8s_rg.name] = k8s_rg



PHIDATA_RUNTIME_ENV_VAR: str = "PHIDATA_RUNTIME"
PYTHONPATH_ENV_VAR: str = "PYTHONPATH"

SCRIPTS_DIR_ENV_VAR: str = "PHI_SCRIPTS_DIR"
STORAGE_DIR_ENV_VAR: str = "PHI_STORAGE_DIR"
META_DIR_ENV_VAR: str = "PHI_META_DIR"
PRODUCTS_DIR_ENV_VAR: str = "PHI_PRODUCTS_DIR"
NOTEBOOKS_DIR_ENV_VAR: str = "PHI_NOTEBOOKS_DIR"
WORKFLOWS_DIR_ENV_VAR: str = "PHI_WORKFLOWS_DIR"
WORKSPACE_ROOT_ENV_VAR: str = "PHI_WORKSPACE_ROOT"
WORKSPACE_MOUNT_ENV_VAR: str = "PHI_WORKSPACE_MOUNT"
WORKSPACE_CONFIG_DIR_ENV_VAR: str = "PHI_WORKSPACE_CONFIG_DIR"
WORKSPACE_CONFIG_FILE_ENV_VAR: str = "PHI_WORKSPACE_CONFIG_FILE"

AWS_REGION_ENV_VAR: str = "AWS_REGION"
AWS_DEFAULT_REGION_ENV_VAR: str = "AWS_DEFAULT_REGION"
AWS_PROFILE_ENV_VAR: str = "AWS_PROFILE"
AWS_CONFIG_FILE_ENV_VAR: str = "AWS_CONFIG_FILE"
AWS_SHARED_CREDENTIALS_FILE_ENV_VAR: str = "AWS_SHARED_CREDENTIALS_FILE"

INIT_AIRFLOW_ENV_VAR: str = "INIT_AIRFLOW"
AIRFLOW_ENV_ENV_VAR: str = "AIRFLOW_ENV"
AIRFLOW_EXECUTOR_ENV_VAR: str = "AIRFLOW__CORE__EXECUTOR"
AIRFLOW_DAGS_FOLDER_ENV_VAR: str = "AIRFLOW__CORE__DAGS_FOLDER"
AIRFLOW_DB_CONN_URL_ENV_VAR: str = "AIRFLOW__DATABASE__SQL_ALCHEMY_CONN"



from typing import Any, Dict, List, Optional

from pydantic import Field

from phidata.infra.base.resource import InfraResource
from phidata.infra.k8s.api_client import K8sApiClient
from phidata.infra.k8s.constants import DEFAULT_K8S_NAMESPACE
from phidata.infra.k8s.enums.api_version import ApiVersion
from phidata.infra.k8s.enums.kind import Kind
from phidata.infra.k8s.resource.meta.v1.object_meta import ObjectMeta
from phidata.utils.cli_console import print_info
from phidata.utils.log import logger


class K8sObject(InfraResource):
    def get_k8s_object(self) -> Any:
        """Returns the K8sObject definition"""
        logger.error("@get_k8s_object method not defined")


class K8sResource(K8sObject):
    """Base class for K8s Resources"""

    ## Common fields for all K8s Resources
    # Which version of the Kubernetes API you're using to create this object
    # Note: we use an alias "apiVersion" so that the K8s manifest generated by this resource
    #       has the correct key
    api_version: ApiVersion = Field(..., alias="apiVersion")
    # What kind of object you want to create
    kind: Kind
    # Data that helps uniquely identify the object, including a name string, UID, and optional namespace
    metadata: ObjectMeta

    ## Fields used in api calls
    # async_req bool: execute request asynchronously
    async_req: bool = False
    # pretty: If 'true', then the output is pretty printed.
    pretty: bool = True

    # List of fields to include from the K8sResource class when generating the
    # K8s manifest. Subclasses must add fields to include in the fields_for_k8s_manifest list
    fields_for_k8s_manifest_base: List[str] = [
        "api_version",
        "apiVersion",
        "kind",
        "metadata",
    ]
    # List of fields to include from Subclasses
    # This should be defined by the Subclass
    fields_for_k8s_manifest: List[str] = []

    # Used for running post-create, post-update and post-delete steps
    resource_available: bool = False
    resource_updated: bool = False
    resource_deleted: bool = False

    def is_valid(self) -> bool:
        # SubResources can use this function to add validation checks
        return True

    def get_namespace(self) -> str:
        if self.metadata and self.metadata.namespace:
            return self.metadata.namespace
        return DEFAULT_K8S_NAMESPACE

    def get_label_selector(self) -> str:
        labels = self.metadata.labels
        if labels:
            label_str = ",".join([f"{k}={v}" for k, v in labels.items()])
            return label_str
        return ""

    def get_k8s_object(self) -> Any:
        """Creates a K8sObject for this resource.
        Eg:
            * For a Deployment resource, it will return the V1Deployment object.
        """
        logger.error("@get_k8s_object method not defined")
        return None

    def get_resource_name(self) -> str:
        if self.name:
            return self.name
        return self.metadata.name or self.__class__.__name__

    @staticmethod
    def get_from_cluster(
        k8s_client: K8sApiClient, namespace: Optional[str] = None, **kwargs
    ) -> Any:
        """Gets all resources of this type from the K8s Cluster

        Args:
            k8s_client: The K8sApiClient for the current Cluster
            namespace: Namespace to use.
        """
        logger.error(f"@get_from_cluster method not defined")
        return None

    def _create(self, k8s_client: K8sApiClient) -> bool:
        logger.error(f"@_create method not defined for {self.__class__.__name__}")
        return False

    def create(self, k8s_client: K8sApiClient) -> bool:
        """Creates the resource on the K8s Cluster

        Args:
            k8s_client: The K8sApiClient for the current Cluster
        """
        if not self.is_valid():
            return False
        if self.use_cache and self.is_active_on_cluster(k8s_client):
            self.resource_available = True
            print_info(
                f"{self.get_resource_type()} {self.get_resource_name()} already active."
            )
            return True
        else:
            self.resource_available = self._create(k8s_client)
        if self.resource_available:
            print_info(
                f"{self.get_resource_type()} {self.get_resource_name()} created."
            )
            return self.post_create(k8s_client)
        return self.resource_available

    def post_create(self, k8s_client: K8sApiClient) -> bool:
        # return True because this function is not used for most resources
        return True

    def _read(self, k8s_client: K8sApiClient) -> Any:
        logger.error(f"@_read method not defined for {self.__class__.__name__}")
        return False

    def read(self, k8s_client: K8sApiClient) -> Any:
        """Reads the resource from the K8s Cluster
        Eg:
            * For a Deployment resource, it will return the V1Deployment object
            currently running on the cluster.

        Args:
            k8s_client: The K8sApiClient for the current Cluster
        """
        if not self.is_valid():
            return None
        if self.use_cache and self.active_resource is not None:
            return self.active_resource
        return self._read(k8s_client)

    def _update(self, k8s_client: K8sApiClient) -> Any:
        logger.error(f"@_update method not defined for {self.__class__.__name__}")
        return False

    def update(self, k8s_client: K8sApiClient) -> bool:
        """Updates the resource on the K8s Cluster

        Args:
            k8s_client: The K8sApiClient for the current Cluster
        """
        if not self.is_valid():
            return False
        if self.is_active_on_cluster(k8s_client):
            self.resource_updated = self._update(k8s_client)
        else:
            print_info(
                f"{self.get_resource_type()} {self.get_resource_name()} does not exist."
            )
            return True
        if self.resource_updated:
            return self.post_update(k8s_client)
        return self.resource_updated

    def post_update(self, k8s_client: K8sApiClient) -> bool:
        # return True because this function is not used for most resources
        return True

    def _delete(self, k8s_client: K8sApiClient) -> Any:
        logger.error(f"@_delete method not defined for {self.__class__.__name__}")
        return False

    def delete(self, k8s_client: K8sApiClient) -> bool:
        """Deletes the resource from the K8s Cluster

        Args:
            k8s_client: The K8sApiClient for the current Cluster
        """
        if not self.is_valid():
            return False
        if self.is_active_on_cluster(k8s_client):
            self.resource_deleted = self._delete(k8s_client)
        else:
            print_info(
                f"{self.get_resource_type()} {self.get_resource_name()} does not exist."
            )
            return True
        if self.resource_deleted:
            return self.post_delete(k8s_client)
        return self.resource_deleted

    def post_delete(self, k8s_client: K8sApiClient) -> bool:
        # return True because this function is not used for most resources
        return True

    def is_active_on_cluster(self, k8s_client: K8sApiClient) -> bool:
        """Returns True if the resource is running on the K8s Cluster"""
        active_resource = self.read(k8s_client=k8s_client)
        if active_resource is not None:
            return True
        return False

    ######################################################
    ## Building the k8s_manifest
    ######################################################

    def get_k8s_manifest_dict(self) -> Optional[Dict[str, Any]]:
        """Returns the K8s Manifest for this Object as a dict"""

        from itertools import chain

        k8s_manifest: Dict[str, Any] = {}
        all_attributes: Dict[str, Any] = self.dict(exclude_defaults=True, by_alias=True)
        # logger.debug("All Attributes: {}".format(all_attributes))
        for attr_name in chain(
            self.fields_for_k8s_manifest_base, self.fields_for_k8s_manifest
        ):
            if attr_name in all_attributes:
                k8s_manifest[attr_name] = all_attributes[attr_name]
        # logger.debug(f"k8s_manifest:\n{k8s_manifest}")
        return k8s_manifest

    def get_k8s_manifest_yaml(self, **kwargs) -> Optional[str]:
        """Returns the K8s Manifest for this Object as a yaml"""
        import yaml

        k8s_manifest_dict = self.get_k8s_manifest_dict()

        if k8s_manifest_dict is not None:
            return yaml.safe_dump(k8s_manifest_dict, **kwargs)
        return None

    def get_k8s_manifest_json(self, **kwargs) -> Optional[str]:
        """Returns the K8s Manifest for this Object as a json"""
        import json

        k8s_manifest_dict = self.get_k8s_manifest_dict()

        if k8s_manifest_dict is not None:
            return json.dumps(k8s_manifest_dict, **kwargs)
        return None



from typing import Optional

from phidata.infra.aws.api_client import AwsApiClient
from phidata.infra.aws.resource.base import AwsResource
from phidata.utils.cli_console import print_error
from phidata.utils.log import logger


class Subnet(AwsResource):
    """
    # https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/ec2.html#subnet
    """

    resource_type = "Subnet"
    service_name = "ec2"

    # Subnet id
    id: str

    def get_resource_name(self) -> Optional[str]:
        return self.id

    def get_availability_zone(
        self, aws_client: Optional[AwsApiClient] = None
    ) -> Optional[str]:
        # logger.debug(f"Reading {self.get_resource_type()}: {self.get_resource_name()}")

        from botocore.exceptions import ClientError

        client: AwsApiClient = aws_client or self.get_aws_client()
        service_resource = self.get_service_resource(client)
        try:
            subnet = service_resource.Subnet(self.id)
            az = subnet.availability_zone
            logger.debug(f"AZ for {self.id}: {az}")
            return az
        except ClientError as ce:
            logger.debug(f"ClientError: {ce}")
        except Exception as e:
            print_error(f"Error reading {self.get_resource_type()}: {e}")
        return None



import enum
from typing import Any, List, Optional


class ExtendedEnum(enum.Enum):
    @classmethod
    def values_list(cls: Any) -> List[Any]:
        return list(map(lambda c: c.value, cls))

    @classmethod
    def from_str(cls: Any, str_to_convert_to_enum: Optional[str]) -> Optional[Any]:
        """Convert a string value to an enum object. Case Sensitive"""

        if str_to_convert_to_enum is None:
            return None

        if str_to_convert_to_enum in cls._value2member_map_:
            return cls._value2member_map_.get(str_to_convert_to_enum)
        else:
            raise NotImplementedError(
                "{} is not a member of {}: {}".format(
                    str_to_convert_to_enum, cls, cls._value2member_map_.keys()
                )
            )



from pathlib import Path
from typing import List, Optional, Dict, Any

from pydantic import validator

from phidata.app import PhidataApp
from phidata.app.databox import default_databox_name
from phidata.infra.base import InfraArgs
from phidata.infra.k8s.resource.group import K8sResourceGroup
from phidata.infra.k8s.create.group import CreateK8sResourceGroup


class K8sArgs(InfraArgs):
    # K8s namespace to use
    namespace: str = "default"
    # K8s context to use
    context: Optional[str] = None
    # K8s service account to use
    service_account_name: Optional[str] = None
    # Common K8s labels to add to all resources
    common_labels: Optional[Dict[str, str]] = None
    # PhidataApp to deploy
    apps: Optional[List[PhidataApp]] = None
    # K8sResourceGroup to deploy
    resources: Optional[List[K8sResourceGroup]] = None
    # CreateK8sResourceGroup to deploy
    create_resources: Optional[List[CreateK8sResourceGroup]] = None
    # Resources dir where k8s manifests are stored
    resources_dir: str = "k8s"
    # databox name for `phi dx ...` commands
    databox_name: str = default_databox_name
    kubeconfig_path: Optional[Path] = None

    # Get K8s context and kubeconfig from an EksCluster resource
    eks_cluster: Optional[Any] = None

    @validator("apps")
    def apps_are_valid(cls, apps):
        if apps is not None:
            for _app in apps:
                if not isinstance(_app, PhidataApp):
                    raise TypeError(f"App not of type PhidataApp: {_app}")
            return apps

    @validator("eks_cluster")
    def eks_cluster_is_valid(cls, eks_cluster):
        if eks_cluster is not None:
            from phidata.infra.aws.resource.eks.cluster import EksCluster

            if not isinstance(eks_cluster, EksCluster):
                raise TypeError("eks_cluster not of type EksCluster")
            return eks_cluster



from typing import Optional, Any, Union, List, Dict
from typing_extensions import Literal

from sqlalchemy.engine import Engine, Connection
from sqlalchemy.exc import ResourceClosedError

from phidata.asset import DataAsset, DataAssetArgs
from phidata.app.db import DbApp
from phidata.task import Task
from phidata.check import Check

from phidata.utils.enums import ExtendedEnum
from phidata.utils.log import logger
from phidata.types.phidata_runtime import PhidataRuntimeType


class SqlType(ExtendedEnum):
    POSTGRES = "POSTGRES"


class SqlTableArgs(DataAssetArgs):
    # Table Name
    name: str
    # Type of SQL table
    sql_type: SqlType
    # Table schema
    db_schema: Optional[str] = None
    # sqlalchemy.engine.(Engine or Connection)
    # Using SQLAlchemy makes it possible to use any DB supported by that library.
    # NOTE: db_engine is required but can be derived using other args.
    db_engine: Optional[Union[Engine, Connection]] = None
    # a db_conn_url can be used to create the sqlalchemy.engine.Engine object
    db_conn_url: Optional[str] = None
    # airflow connection_id used for running workflows on airflow
    airflow_conn_id: Optional[str] = None
    # Phidata DbApp to connect to the database
    db_app: Optional[DbApp] = None

    cached_db_engine: Optional[Union[Engine, Connection]] = None


class SqlTable(DataAsset):
    """Base Class for Sql tables"""

    def __init__(
        self,
        # Table Name,
        name: Optional[str] = None,
        # Type of SQL table,
        sql_type: Optional[SqlType] = None,
        # Table schema,
        db_schema: Optional[str] = None,
        # sqlalchemy.engine.(Engine or Connection),
        # Using SQLAlchemy makes it possible to use any DB supported by that library.,
        # NOTE: db_engine is required but can be derived using other args.,
        db_engine: Optional[Union[Engine, Connection]] = None,
        # a db_conn_url can be used to create the sqlalchemy.engine.Engine object,
        db_conn_url: Optional[str] = None,
        # airflow connection_id used for running workflows on airflow,
        airflow_conn_id: Optional[str] = None,
        # Phidata DbApp to connect to the database,
        db_app: Optional[DbApp] = None,
        # Checks to run before loading the table,
        pre_checks: Optional[List[Check]] = None,
        # List of tasks to create the table
        create_tasks: Optional[List[Task]] = None,
        # Checks to run after loading the table,
        post_checks: Optional[List[Check]] = None,
        # List of tasks to update the table,
        update_tasks: Optional[List[Task]] = None,
        # List of tasks to delete the table,
        delete_tasks: Optional[List[Task]] = None,
        # Control which environment the table is created in
        env: Optional[str] = None,
        # Dev Args,
        dev_name: Optional[str] = None,
        dev_env: Optional[Dict[str, Any]] = None,
        seed_dev_tasks: Optional[List[Task]] = None,
        dev_stg_swap_tasks: Optional[List[Task]] = None,
        # Staging Args,
        stg_name: Optional[str] = None,
        stg_env: Optional[Dict[str, Any]] = None,
        seed_stg_tasks: Optional[List[Task]] = None,
        stg_prd_swap_tasks: Optional[List[Task]] = None,
        # Production Args,
        prd_name: Optional[str] = None,
        prd_env: Optional[Dict[str, Any]] = None,
        cached_db_engine: Optional[Union[Engine, Connection]] = None,
        version: Optional[str] = None,
        enabled: bool = True,
        **kwargs,
    ) -> None:
        super().__init__()
        self.args: Optional[SqlTableArgs] = None
        if name is not None and sql_type is not None:
            try:
                self.args = SqlTableArgs(
                    name=name,
                    sql_type=sql_type,
                    db_schema=db_schema,
                    db_engine=db_engine,
                    db_conn_url=db_conn_url,
                    airflow_conn_id=airflow_conn_id,
                    db_app=db_app,
                    pre_checks=pre_checks,
                    create_tasks=create_tasks,
                    post_checks=post_checks,
                    update_tasks=update_tasks,
                    delete_tasks=delete_tasks,
                    env=env,
                    dev_name=dev_name,
                    dev_env=dev_env,
                    seed_dev_tasks=seed_dev_tasks,
                    dev_stg_swap_tasks=dev_stg_swap_tasks,
                    stg_name=stg_name,
                    stg_env=stg_env,
                    seed_stg_tasks=seed_stg_tasks,
                    stg_prd_swap_tasks=stg_prd_swap_tasks,
                    prd_name=prd_name,
                    prd_env=prd_env,
                    cached_db_engine=cached_db_engine,
                    version=version,
                    enabled=enabled,
                    **kwargs,
                )
            except Exception as e:
                logger.error(f"Args for {self.name} are not valid")
                raise

    @property
    def sql_type(self) -> Optional[SqlType]:
        return self.args.sql_type if self.args else None

    @sql_type.setter
    def sql_type(self, sql_type: SqlType) -> None:
        if self.args and sql_type:
            self.args.sql_type = sql_type

    @property
    def db_schema(self) -> Optional[str]:
        return self.args.db_schema if self.args else None

    @db_schema.setter
    def db_schema(self, db_schema: str) -> None:
        if self.args and db_schema:
            self.args.db_schema = db_schema

    @property
    def db_engine(self) -> Optional[Union[Engine, Connection]]:
        return self.args.db_engine if self.args else None

    @db_engine.setter
    def db_engine(self, db_engine: Union[Engine, Connection]) -> None:
        if self.args and db_engine:
            self.args.db_engine = db_engine

    @property
    def db_conn_url(self) -> Optional[str]:
        return self.args.db_conn_url if self.args else None

    @db_conn_url.setter
    def db_conn_url(self, db_conn_url: str) -> None:
        if self.args and db_conn_url:
            self.args.db_conn_url = db_conn_url

    @property
    def airflow_conn_id(self) -> Optional[str]:
        return self.args.airflow_conn_id if self.args else None

    @airflow_conn_id.setter
    def airflow_conn_id(self, airflow_conn_id: str) -> None:
        if self.args and airflow_conn_id:
            self.args.airflow_conn_id = airflow_conn_id

    @property
    def db_app(self) -> Optional[DbApp]:
        return self.args.db_app if self.args else None

    @db_app.setter
    def db_app(self, db_app: DbApp) -> None:
        if self.args and db_app:
            self.args.db_app = db_app

    @property
    def cached_db_engine(self) -> Optional[Union[Engine, Connection]]:
        return self.args.cached_db_engine if self.args else None

    @cached_db_engine.setter
    def cached_db_engine(self, cached_db_engine: Union[Engine, Connection]) -> None:
        if self.args and cached_db_engine:
            self.args.cached_db_engine = cached_db_engine

    def create_db_engine_using_conn_url(self) -> Optional[Union[Engine, Connection]]:
        # Create the SQLAlchemy engine using db_conn_url

        try:
            from sqlalchemy import create_engine

            # logger.info(f"Creating db_engine using db_conn_url: {self.db_conn_url}")
            db_engine = create_engine(self.db_conn_url)
            logger.debug(f"db_engine: {db_engine}")
            if isinstance(db_engine, tuple) and len(db_engine) > 0:
                self.db_engine = db_engine[0]
            else:
                self.db_engine = db_engine
            return self.db_engine
        except Exception as e:
            logger.error(f"Error creating db_engine using {self.db_conn_url}")
            logger.error(e)
            return None

    def create_db_engine_using_airflow_conn_id(
        self,
    ) -> Optional[Union[Engine, Connection]]:
        # Create the SQLAlchemy engine using airflow_conn_id

        try:
            from airflow.providers.postgres.hooks.postgres import PostgresHook

            logger.info(
                f"Creating db_engine using airflow_conn_id: {self.airflow_conn_id}"
            )
            if self.sql_type == SqlType.POSTGRES:
                pg_hook = PostgresHook(postgres_conn_id=self.airflow_conn_id)
                self.db_engine = pg_hook.get_sqlalchemy_engine()
            return self.db_engine
        except Exception as e:
            logger.error(f"Error creating db_engine using {self.airflow_conn_id}")
            logger.error(e)
            return None

    def create_db_engine_using_db_app(self) -> Optional[Union[Engine, Connection]]:
        # Create the SQLAlchemy engine using db_app

        if self.db_app is None:
            return None

        try:
            conn_url: Optional[str] = None

            phidata_runtime: Optional[PhidataRuntimeType] = self.phidata_runtime
            if phidata_runtime is None:
                conn_url = self.db_app.get_db_connection_url_local()
            if phidata_runtime == PhidataRuntimeType.local:
                conn_url = self.db_app.get_db_connection_url_local()
            if phidata_runtime == PhidataRuntimeType.docker:
                conn_url = self.db_app.get_db_connection_url_docker()
            if phidata_runtime == PhidataRuntimeType.kubernetes:
                conn_url = self.db_app.get_db_connection_url_k8s()

            if conn_url is None or "None" in conn_url:
                return None

            # Create the SQLAlchemy engine using conn_url
            from sqlalchemy import create_engine

            # logger.info(f"Creating db_engine using conn_url: {conn_url}")
            db_engine = create_engine(conn_url)
            logger.debug(f"db_engine: {db_engine}")
            if isinstance(db_engine, tuple) and len(db_engine) > 0:
                self.db_engine = db_engine[0]
            else:
                self.db_engine = db_engine
            return self.db_engine
        except Exception as e:
            logger.error(f"Error creating db_engine using {self.db_app}")
            logger.error(e)
            return None

    def create_db_engine(self) -> Optional[Union[Engine, Connection]]:
        if self.db_engine is not None:
            return self.db_engine

        if self.cached_db_engine is not None:
            return self.cached_db_engine

        if self.db_conn_url is not None:
            conn_url_db_engine = self.create_db_engine_using_conn_url()
            if conn_url_db_engine is not None:
                self.cached_db_engine = conn_url_db_engine
                return conn_url_db_engine

        if self.airflow_conn_id is not None:
            from phidata.airflow.airflow_installed import airflow_installed

            # Validate that airflow is installed on the machine
            if airflow_installed():
                airflow_conn_id_db_engine = (
                    self.create_db_engine_using_airflow_conn_id()
                )
                if airflow_conn_id_db_engine is not None:
                    self.cached_db_engine = airflow_conn_id_db_engine
                    return airflow_conn_id_db_engine

        if self.db_app is not None:
            db_app_db_engine = self.create_db_engine_using_db_app()
            if db_app_db_engine is not None:
                self.cached_db_engine = db_app_db_engine
                return db_app_db_engine

        return None

    ######################################################
    ## Write table
    ######################################################

    def write_pandas_df(
        self,
        df: Optional[Any] = None,
        # How to behave if the table already exists.
        # fail: Raise a ValueError.
        # replace: Drop the table before inserting new values.
        # append: Insert new values to the existing table.
        if_exists: Optional[Literal["fail", "replace", "append"]] = None,
        # Write DataFrame index as a column.
        # Uses index_label as the column name in the table.
        index: Optional[bool] = None,
        # Column label for index column(s).
        # If None is given (default) and index is True, then the index names are used.
        # A sequence should be given if the DataFrame uses MultiIndex.
        index_label: Optional[Union[str, List[str]]] = None,
        # Specify the number of rows in each batch to be written at a time.
        # By default, all rows will be written at once.
        chunksize: Optional[int] = None,
        # Create database if it does not exist.
        create_database: bool = False,
    ) -> bool:
        """
        Write DataFrame to table.
        """

        # SqlTable not yet initialized
        if self.args is None:
            return False

        # Check name is available
        if self.name is None:
            logger.error("SqlTable name not available")
            return False

        # Check engine is available
        db_engine = self.create_db_engine()
        if db_engine is None:
            logger.error("DbEngine not available")
            return False

        # write to table
        import pandas as pd

        if df is None or not isinstance(df, pd.DataFrame):
            logger.error("DataFrame invalid")
            return False

        rows_in_df = df.shape[0]
        logger.info(f"Writing {rows_in_df} rows to table: {self.name}")

        # create a dict of args provided
        not_null_args: Dict[str, Any] = {}
        if self.db_schema:
            not_null_args["schema"] = self.db_schema
        if if_exists:
            not_null_args["if_exists"] = if_exists
        if index:
            not_null_args["index"] = index
        if index_label:
            not_null_args["index_label"] = index_label
        if chunksize:
            not_null_args["chunksize"] = chunksize

        try:
            with db_engine.connect() as connection:
                if self.db_schema is not None and create_database:
                    logger.info(f"Creating database: {self.db_schema}")
                    # Create database if it does not exist
                    connection.execute(f"CREATE SCHEMA IF NOT EXISTS {self.db_schema}")

                df.to_sql(
                    name=self.name,
                    con=connection,
                    **not_null_args,
                )
                logger.info(f"--**-- Done")
            return True
        except Exception:
            logger.error("Could not write table: {}".format(self.name))
            raise

    ######################################################
    ## Read table
    ######################################################

    def read_pandas_df(
        self,
        index_col: Optional[Union[str, List[str]]] = None,
        coerce_float: bool = True,
        parse_dates: Optional[Union[List, Dict]] = None,
        columns: Optional[List[str]] = None,
        chunksize: Optional[int] = None,
    ) -> Optional[Any]:
        """
        Read table into a DataFrame.

        Args:
            index_col : str or list of str, optional, default: None
                Column(s) to set as index(MultiIndex).
            coerce_float : bool, default True
                Attempts to convert values of non-string, non-numeric objects (like
                decimal.Decimal) to floating point. Can result in loss of Precision.
            parse_dates : list or dict, default None
                - List of column names to parse as dates.
                - Dict of ``{column_name: format string}`` where format string is
                strftime compatible in case of parsing string times or is one of
                (D, s, ns, ms, us) in case of parsing integer timestamps.
                - Dict of ``{column_name: arg dict}``, where the arg dict corresponds
                to the keyword arguments of :func:`pandas.to_datetime`
                Especially useful with databases without native Datetime support,
                such as SQLite.
            columns : list, default None
                List of column names to select from SQL table.
            chunksize : int, default None
                If specified, returns an iterator where `chunksize` is the number of
                rows to include in each chunk.

        Returns:
            DataFrame or Iterator[DataFrame]
            A SQL table is returned as two-dimensional data structure with labeled
            axes.
        """

        # SqlTable not yet initialized
        if self.args is None:
            return False

        # Check engine is available
        db_engine = self.create_db_engine()
        if db_engine is None:
            logger.error("DbEngine not available")
            return False

        # read sql table
        import pandas as pd

        logger.info("Reading table: {}".format(self.name))

        # create a dict of args provided
        not_null_args: Dict[str, Any] = {}
        if self.db_schema:
            not_null_args["schema"] = self.db_schema
        if index_col:
            not_null_args["index_col"] = index_col
        if coerce_float:
            not_null_args["coerce_float"] = coerce_float
        if parse_dates:
            not_null_args["parse_dates"] = parse_dates
        if columns:
            not_null_args["columns"] = columns
        if chunksize:
            not_null_args["chunksize"] = chunksize

        try:
            with db_engine.connect() as connection:
                result_df = pd.read_sql_table(
                    table_name=self.name,
                    con=connection,
                    **not_null_args,
                )
            return result_df
        except Exception:
            logger.error(f"Could not read table: {self.name}")
            raise

    def run_sql_query(
        self,
        sql_query: str,
        index_col: Optional[Union[str, List[str]]] = None,
        coerce_float: bool = True,
        parse_dates: Optional[Union[List, Dict]] = None,
        columns: Optional[List[str]] = None,
        chunksize: Optional[int] = None,
    ) -> Optional[Any]:
        """
        Run SQL query using pandas.read_sql()

        Args:
            index_col : str or list of str, optional, default: None
                Column(s) to set as index(MultiIndex).
            coerce_float : bool, default True
                Attempts to convert values of non-string, non-numeric objects (like
                decimal.Decimal) to floating point. Can result in loss of Precision.
            parse_dates : list or dict, default None
                - List of column names to parse as dates.
                - Dict of ``{column_name: format string}`` where format string is
                strftime compatible in case of parsing string times or is one of
                (D, s, ns, ms, us) in case of parsing integer timestamps.
                - Dict of ``{column_name: arg dict}``, where the arg dict corresponds
                to the keyword arguments of :func:`pandas.to_datetime`
                Especially useful with databases without native Datetime support,
                such as SQLite.
            columns : list, default None
                List of column names to select from SQL table.
            chunksize : int, default None
                If specified, returns an iterator where `chunksize` is the number of
                rows to include in each chunk.

        Returns:
            DataFrame or Iterator[DataFrame]
            A SQL table is returned as two-dimensional data structure with labeled
            axes.
        """

        # SqlTable not yet initialized
        if self.args is None:
            return None

        # Check engine is available
        db_engine = self.create_db_engine()
        if db_engine is None:
            logger.error("DbEngine not available")
            return False

        # run sql query
        import pandas as pd

        logger.info("Running Query:\n{}".format(sql_query))

        # create a dict of args provided
        not_null_args: Dict[str, Any] = {}
        if self.db_schema:
            not_null_args["schema"] = self.db_schema
        if index_col:
            not_null_args["index_col"] = index_col
        if coerce_float:
            not_null_args["coerce_float"] = coerce_float
        if parse_dates:
            not_null_args["parse_dates"] = parse_dates
        if columns:
            not_null_args["columns"] = columns
        if chunksize:
            not_null_args["chunksize"] = chunksize

        try:
            with db_engine.connect() as connection:
                result_df = pd.read_sql(
                    sql=sql_query,
                    con=connection,
                    **not_null_args,
                )
            return result_df
        except ResourceClosedError as rce:
            logger.info(
                f"The result object was closed automatically, returning no rows."
            )
        # except Exception as e:
        #     logger.error(f"Sql query failed: {e}")
        return None

    ######################################################
    ## Drop table
    ######################################################

    def _delete(self) -> bool:
        try:
            result = self.run_sql_query(f"DROP TABLE {self.name};")
        except Exception as e:
            logger.error(f"Drop table failed: {e}")
        return True



from typing import Dict, List, Optional

from phidata.infra.k8s.enums.api_group import ApiGroup
from phidata.infra.k8s.enums.api_version import ApiVersion
from phidata.infra.k8s.enums.kind import Kind
from phidata.infra.k8s.resource.rbac_authorization_k8s_io.v1.cluste_role_binding import (
    Subject,
    RoleRef,
    ClusterRoleBinding,
)
from phidata.infra.k8s.create.common.labels import create_component_labels_dict
from phidata.infra.k8s.resource.meta.v1.object_meta import ObjectMeta
from phidata.utils.log import logger


def create_eks_admin_crb(
    name: str = "eks-admin-crb",
    cluster_role: str = "cluster-admin",
    users: Optional[List[str]] = None,
    groups: Optional[List[str]] = None,
    service_accounts: Optional[List[str]] = None,
    app_name: str = "eks-admin",
    labels: Optional[Dict[str, str]] = None,
    skip_create: bool = False,
    skip_delete: bool = False,
) -> Optional[ClusterRoleBinding]:

    crb_labels = create_component_labels_dict(
        component_name=name,
        app_name=app_name,
        labels=labels,
    )

    subjects: List[Subject] = []
    if service_accounts is not None and isinstance(service_accounts, list):
        for sa in service_accounts:
            subjects.append(Subject(kind=Kind.SERVICEACCOUNT, name=sa))
    if users is not None and isinstance(users, list):
        for user in users:
            subjects.append(Subject(kind=Kind.USER, name=user))
    if groups is not None and isinstance(groups, list):
        for group in groups:
            subjects.append(Subject(kind=Kind.GROUP, name=group))

    if len(subjects) == 0:
        logger.error(f"No subjects for ClusterRoleBinding: {name}")
        return None

    return ClusterRoleBinding(
        api_version=ApiVersion.RBAC_AUTH_V1,
        kind=Kind.CLUSTERROLEBINDING,
        metadata=ObjectMeta(
            name=name,
            labels=crb_labels,
        ),
        role_ref=RoleRef(
            api_group=ApiGroup.RBAC_AUTH,
            kind=Kind.CLUSTERROLE,
            name=cluster_role,
        ),
        subjects=subjects,
        skip_create=skip_create,
        skip_delete=skip_delete,
    )



from typing import Optional, Literal, List, Tuple, Dict, Any, Union

from phidata.asset.aws.s3.dataset_base import (
    S3DatasetBase,
    S3DatasetBaseArgs,
    S3DatasetType,
)
from phidata.task import Task
from phidata.check import Check
from phidata.infra.aws.resource.s3.bucket import S3Bucket
from phidata.utils.log import logger


class S3DatasetArgs(S3DatasetBaseArgs):
    # Compression style (None, snappy, gzip, zstd).
    compression: Optional[str] = None
    # Max number of rows in each file.
    # Default is None i.e. do not split the files.
    # (e.g. 33554432, 268435456)
    max_rows_by_file: Optional[int] = None
    # Additional parameters forwarded to pyarrow.
    # e.g. pyarrow_additional_kwargs={‘coerce_timestamps’: ‘ns’, ‘allow_truncated_timestamps’=False}
    pyarrow_additional_kwargs: Optional[Dict[str, Any]] = None


class S3Dataset(S3DatasetBase):
    def __init__(
        self,
        table: str,
        database: Optional[str] = None,
        table_type: Optional[str] = None,
        table_description: Optional[str] = None,
        write_mode: Optional[
            Literal["append", "overwrite", "overwrite_partitions"]
        ] = None,
        path: Optional[str] = None,
        top_level_dir: Optional[str] = "datasets",
        path_prefix: Optional[str] = None,
        bucket: Optional[S3Bucket] = None,
        filename_prefix: Optional[str] = None,
        # Checks to run before loading the table,
        pre_checks: Optional[List[Check]] = None,
        # List of tasks to create the table
        create_tasks: Optional[List[Task]] = None,
        # Checks to run after loading the table,
        post_checks: Optional[List[Check]] = None,
        # List of tasks to update the table,
        update_tasks: Optional[List[Task]] = None,
        # List of tasks to delete the table,
        delete_tasks: Optional[List[Task]] = None,
        env: Optional[str] = None,
        # Dev Args,
        dev_name: Optional[str] = None,
        dev_env: Optional[Dict[str, Any]] = None,
        seed_dev_tasks: Optional[List[Task]] = None,
        dev_stg_swap_tasks: Optional[List[Task]] = None,
        # Staging Args,
        stg_name: Optional[str] = None,
        stg_env: Optional[Dict[str, Any]] = None,
        seed_stg_tasks: Optional[List[Task]] = None,
        stg_prd_swap_tasks: Optional[List[Task]] = None,
        # Production Args,
        prd_name: Optional[str] = None,
        prd_env: Optional[Dict[str, Any]] = None,
        dtype: Optional[Dict[str, str]] = None,
        parameters: Optional[Dict[str, str]] = None,
        columns_comments: Optional[Dict[str, str]] = None,
        compression: Optional[str] = None,
        max_rows_by_file: Optional[int] = None,
        pyarrow_additional_kwargs: Optional[Dict[str, Any]] = None,
        partition_cols: Optional[List[str]] = None,
        bucketing_info: Optional[Tuple[List[str], int]] = None,
        concurrent_partitioning: Optional[bool] = None,
        catalog_versioning: Optional[bool] = None,
        schema_evolution: Optional[bool] = None,
        catalog_id: Optional[str] = None,
        use_threads: Optional[Union[bool, int]] = None,
        boto3_session: Optional[Any] = None,
        s3_additional_kwargs: Optional[Dict[str, Any]] = None,
        regular_partitions: Optional[bool] = None,
        projection_enabled: Optional[bool] = None,
        projection_types: Optional[Dict[str, str]] = None,
        projection_ranges: Optional[Dict[str, str]] = None,
        projection_values: Optional[Dict[str, str]] = None,
        projection_intervals: Optional[Dict[str, str]] = None,
        projection_digits: Optional[Dict[str, str]] = None,
        projection_formats: Optional[Dict[str, str]] = None,
        projection_storage_location_template: Optional[str] = None,
        version: Optional[str] = None,
        enabled: bool = True,
        **kwargs,
    ) -> None:

        super().__init__()
        try:
            self.args: S3DatasetArgs = S3DatasetArgs(
                name=f"{database}.{table}",
                dataset_type=S3DatasetType.PARQUET,
                table=table,
                database=database,
                table_type=table_type,
                table_description=table_description,
                write_mode=write_mode,
                path=path,
                top_level_dir=top_level_dir,
                path_prefix=path_prefix,
                bucket=bucket,
                filename_prefix=filename_prefix,
                pre_checks=pre_checks,
                create_tasks=create_tasks,
                post_checks=post_checks,
                update_tasks=update_tasks,
                delete_tasks=delete_tasks,
                env=env,
                dev_name=dev_name,
                dev_env=dev_env,
                seed_dev_tasks=seed_dev_tasks,
                dev_stg_swap_tasks=dev_stg_swap_tasks,
                stg_name=stg_name,
                stg_env=stg_env,
                seed_stg_tasks=seed_stg_tasks,
                stg_prd_swap_tasks=stg_prd_swap_tasks,
                prd_name=prd_name,
                prd_env=prd_env,
                dtype=dtype,
                parameters=parameters,
                columns_comments=columns_comments,
                compression=compression,
                max_rows_by_file=max_rows_by_file,
                pyarrow_additional_kwargs=pyarrow_additional_kwargs,
                partition_cols=partition_cols,
                bucketing_info=bucketing_info,
                concurrent_partitioning=concurrent_partitioning,
                catalog_versioning=catalog_versioning,
                schema_evolution=schema_evolution,
                catalog_id=catalog_id,
                use_threads=use_threads,
                boto3_session=boto3_session,
                s3_additional_kwargs=s3_additional_kwargs,
                regular_partitions=regular_partitions,
                projection_enabled=projection_enabled,
                projection_types=projection_types,
                projection_ranges=projection_ranges,
                projection_values=projection_values,
                projection_intervals=projection_intervals,
                projection_digits=projection_digits,
                projection_formats=projection_formats,
                projection_storage_location_template=projection_storage_location_template,
                version=version,
                enabled=enabled,
                **kwargs,
            )
        except Exception as e:
            logger.error(f"Args for {self.name} are not valid")
            raise

    @property
    def compression(self) -> Optional[str]:
        return self.args.compression

    @compression.setter
    def compression(self, compression: str) -> None:
        if compression is not None:
            self.args.compression = compression

    @property
    def max_rows_by_file(self) -> Optional[int]:
        return self.args.max_rows_by_file

    @max_rows_by_file.setter
    def max_rows_by_file(self, max_rows_by_file: int) -> None:
        if max_rows_by_file is not None:
            self.args.max_rows_by_file = max_rows_by_file

    @property
    def pyarrow_additional_kwargs(self) -> Optional[Dict[str, Any]]:
        return self.args.pyarrow_additional_kwargs

    @pyarrow_additional_kwargs.setter
    def pyarrow_additional_kwargs(
        self, pyarrow_additional_kwargs: Dict[str, Any]
    ) -> None:
        if pyarrow_additional_kwargs is not None:
            self.args.pyarrow_additional_kwargs = pyarrow_additional_kwargs

    ######################################################
    ## Write dataset
    ######################################################

    def write_pandas_df(
        self,
        df: Optional[Any] = None,
        # S3 path for the dataset or use the path defined in the dataset
        path: Optional[str] = None,
        # True to store the DataFrame index in file, otherwise False to ignore it.
        index: Optional[bool] = None,
        # Compression style (None, snappy, gzip, zstd).,
        compression: Optional[str] = None,
        # Max number of rows in each file.,
        # Default is None i.e. dont split the files. (e.g. 33554432, 268435456),
        max_rows_by_file: Optional[int] = None,
        # Additional parameters forwarded to pyarrow.,
        # e.g. pyarrow_additional_kwargs={‘coerce_timestamps’: ‘ns’, ‘allow_truncated_timestamps’=False},
        pyarrow_additional_kwargs: Optional[Dict[str, Any]] = None,
        # True to enable concurrent requests, False to disable multiple threads.,
        # If enabled os.cpu_count() will be used as the max number of threads.,
        # If integer is provided, specified number is used.,
        use_threads: Optional[Union[bool, int]] = None,
        # Boto3 Session. The default boto3 Session will be used if boto3_session receive None.,
        boto3_session: Optional[Any] = None,
        # Forwarded to botocore requests.,
        # e.g. s3_additional_kwargs={‘ServerSideEncryption’: ‘aws:kms’, ‘SSEKMSKeyId’: ‘YOUR_KMS_KEY_ARN’},
        s3_additional_kwargs: Optional[Dict[str, Any]] = None,
        # Add a filename prefix to the output files.
        filename_prefix: Optional[str] = None,
        # List of column names that will be used to create partitions.,
        partition_cols: Optional[List[str]] = None,
        # Tuple consisting of the column names used for,
        # bucketing as the first element and the number of buckets as the second element.,
        # Only str, int and bool are supported as column data types for bucketing.,
        bucketing_info: Optional[Tuple[List[str], int]] = None,
        # If True will increase the parallelism level during the partitions writing.,
        # It will decrease the writing time and increase the memory usage.,
        # https://aws-sdk-pandas.readthedocs.io/en/2.17.0/tutorials/022%20-%20Writing%20Partitions%20Concurrently.html,
        concurrent_partitioning: Optional[bool] = None,
        # write mode – append (Default), overwrite, overwrite_partitions.,
        write_mode: Optional[
            Literal["append", "overwrite", "overwrite_partitions"]
        ] = None,
        # If True and mode=”overwrite”, creates an archived version of the table catalog before updating it.,
        catalog_versioning: Optional[bool] = None,
        # If True allows schema evolution (new or missing columns), otherwise a exception will be raised.,
        # True by default. (Only considered if mode in (“append”, “overwrite_partitions”)),
        schema_evolution: Optional[bool] = None,
        # Glue/Athena catalog: Database name.,
        database: Optional[str] = None,
        # Create Glue/Athena database if it does not exist.
        create_database: bool = False,
        # Glue/Athena catalog: Table name.,
        table: Optional[str] = None,
        # The type of the Glue Table. Set to EXTERNAL_TABLE if None.,
        table_type: Optional[str] = None,
        # Dictionary of columns names and Athena/Glue types to be casted.,
        # Useful when you have columns with undetermined or mixed data types.,
        # (e.g. {‘col name’: ‘bigint’, ‘col2 name’: ‘int’}),
        dtype: Optional[Dict[str, str]] = None,
        # Glue/Athena catalog: Table description,
        table_description: Optional[str] = None,
        # Glue/Athena catalog: Key/value pairs to tag the table.,
        parameters: Optional[Dict[str, str]] = None,
        # Glue/Athena catalog: Columns names and the related comments,
        # (e.g. {‘col0’: ‘Column 0.’, ‘col1’: ‘Column 1.’, ‘col2’: ‘Partition.’}).,
        columns_comments: Optional[Dict[str, str]] = None,
        # regular_partitions (bool) – Create regular partitions (Non projected partitions) on Glue Catalog.,
        # Disable when you will work only with Partition Projection.,
        regular_partitions: Optional[bool] = None,
        # Enable Partition Projection on Athena,
        # https://docs.aws.amazon.com/athena/latest/ug/partition-projection.html,
        projection_enabled: Optional[bool] = None,
        # Dictionary of partitions names and Athena projections types. Valid types: “enum”, “integer”, “date”, “injected”,
        # https://docs.aws.amazon.com/athena/latest/ug/partition-projection-supported-types.html,
        # e.g. {‘col_name’: ‘enum’, ‘col2_name’: ‘integer’},
        projection_types: Optional[Dict[str, str]] = None,
        # Dictionary of partitions names and Athena projections ranges.,
        # https://docs.aws.amazon.com/athena/latest/ug/partition-projection-supported-types.html,
        # e.g. {‘col_name’: ‘0,10’, ‘col2_name’: ‘-1,8675309’},
        projection_ranges: Optional[Dict[str, str]] = None,
        # Dictionary of partitions names and Athena projections values.,
        # https://docs.aws.amazon.com/athena/latest/ug/partition-projection-supported-types.html,
        # e.g. {‘col_name’: ‘A,B,Unknown’, ‘col2_name’: ‘foo,boo,bar’},
        projection_values: Optional[Dict[str, str]] = None,
        # Dictionary of partitions names and Athena projections intervals.,
        # https://docs.aws.amazon.com/athena/latest/ug/partition-projection-supported-types.html,
        # e.g. {‘col_name’: ‘1’, ‘col2_name’: ‘5’},
        projection_intervals: Optional[Dict[str, str]] = None,
        # Dictionary of partitions names and Athena projections digits.,
        # https://docs.aws.amazon.com/athena/latest/ug/partition-projection-supported-types.html,
        # e.g. {‘col_name’: ‘1’, ‘col2_name’: ‘2’},
        projection_digits: Optional[Dict[str, str]] = None,
        # The ID of the Data Catalog from which to retrieve Databases.,
        # If none is provided, the AWS account ID is used by default.,
        catalog_id: Optional[str] = None,
    ) -> bool:
        """
        Write DataFrame to S3Dataset.
        """

        # S3Dataset not yet initialized
        if self.args is None:
            return False

        if df is None:
            logger.error("DataFrame not provided")
            return False

        dataset_path = path or self.uri
        if dataset_path is None:
            logger.error("Dataset path not provided")
            return False
        logger.info(f"Writing to {dataset_path}")

        # create a dict of args which are not null
        not_null_args: Dict[str, Any] = {}

        if index is not None:
            not_null_args["index"] = index

        if compression is not None:
            not_null_args["compression"] = compression

        if max_rows_by_file is not None:
            not_null_args["max_rows_by_file"] = max_rows_by_file

        if pyarrow_additional_kwargs is not None:
            not_null_args["pyarrow_additional_kwargs"] = pyarrow_additional_kwargs

        _table = table or self.table
        if _table is not None:
            not_null_args["table"] = _table

        _database = database or self.database
        if _database is not None:
            not_null_args["database"] = _database

        _table_type = table_type or self.table_type
        if _table_type is not None:
            not_null_args["table_type"] = _table_type

        _table_description = table_description or self.table_description
        if _table_description is not None:
            not_null_args["description"] = _table_description

        _write_mode = write_mode or self.write_mode
        if _write_mode is not None:
            not_null_args["mode"] = _write_mode

        _dtype = dtype or self.dtype
        if _dtype is not None:
            not_null_args["dtype"] = _dtype

        _parameters = parameters or self.parameters
        if _parameters is not None:
            not_null_args["parameters"] = _parameters

        _columns_comments = columns_comments or self.columns_comments
        if _columns_comments is not None:
            not_null_args["columns_comments"] = _columns_comments

        _filename_prefix = filename_prefix or self.filename_prefix
        if _filename_prefix is not None:
            not_null_args["filename_prefix"] = _filename_prefix

        _partition_cols = partition_cols or self.partition_cols
        if _partition_cols is not None:
            not_null_args["partition_cols"] = _partition_cols

        _bucketing_info = bucketing_info or self.bucketing_info
        if _bucketing_info is not None:
            not_null_args["bucketing_info"] = _bucketing_info

        _concurrent_partitioning = (
            concurrent_partitioning or self.concurrent_partitioning
        )
        if _concurrent_partitioning is not None:
            not_null_args["concurrent_partitioning"] = _concurrent_partitioning

        _catalog_versioning = catalog_versioning or self.catalog_versioning
        if _catalog_versioning is not None:
            not_null_args["catalog_versioning"] = _catalog_versioning

        _schema_evolution = schema_evolution or self.schema_evolution
        if _schema_evolution is not None:
            not_null_args["schema_evolution"] = _schema_evolution

        _catalog_id = catalog_id or self.catalog_id
        if _catalog_id is not None:
            not_null_args["catalog_id"] = _catalog_id

        _use_threads = use_threads or self.use_threads
        if _use_threads is not None:
            not_null_args["use_threads"] = _use_threads

        _s3_additional_kwargs = s3_additional_kwargs or self.s3_additional_kwargs
        if _s3_additional_kwargs is not None:
            not_null_args["s3_additional_kwargs"] = _s3_additional_kwargs

        _regular_partitions = regular_partitions  # or self.regular_partitionsl
        if _regular_partitions is not None:
            not_null_args["regular_partitions"] = _regular_partitions

        _projection_enabled = projection_enabled  # or self.projection_enabled
        if _projection_enabled is not None:
            not_null_args["projection_enabled"] = _projection_enabled

        _projection_types = projection_types  # or self.projection_types
        if _projection_types is not None:
            not_null_args["projection_types"] = _projection_types

        _projection_ranges = projection_ranges  # or self.projection_ranges
        if _projection_ranges is not None:
            not_null_args["projection_ranges"] = _projection_ranges

        _projection_values = projection_values  # or self.projection_values
        if _projection_values is not None:
            not_null_args["projection_values"] = _projection_values

        _projection_intervals = projection_intervals  # or self.projection_intervals
        if _projection_intervals is not None:
            not_null_args["projection_intervals"] = _projection_intervals

        _projection_digits = projection_digits  # or self.projection_digits
        if _projection_digits is not None:
            not_null_args["projection_digits"] = _projection_digits

        try:
            import awswrangler as wr

            # Create boto3 session
            _boto3_session = boto3_session or self.boto3_session
            if _boto3_session is None:
                logger.info("Creating boto3 session using aws_api_client")
                aws_api_client = self.aws_api_client
                if aws_api_client is not None:
                    _boto3_session = aws_api_client.boto3_session

            # Create database if needed
            if _database is not None and create_database:
                logger.info(f"Creating database: '{_database}'")
                wr.catalog.create_database(
                    name=_database,
                    exist_ok=True,
                    catalog_id=_catalog_id,
                    boto3_session=_boto3_session,
                )

            response: Dict[
                str, Union[List[str], Dict[str, List[str]]]
            ] = wr.s3.to_parquet(
                df=df,
                path=dataset_path,
                dataset=True,
                sanitize_columns=True,
                boto3_session=_boto3_session,
                **not_null_args,
            )
            logger.info(f"Dataset {self.name} written to {dataset_path}")
            logger.debug(f"Response: {response}")
            logger.info("--**-- Done")
            return True
        except Exception as e:
            logger.error("Could not write to dataset: {}".format(self.name))
            raise

    def create_from_query(
        self,
        sql: str,
        # If creating from a table stored in a different database
        # origin_database is the database where the original table is stored.
        origin_database: Optional[str] = None,
        # S3 path for the dataset or use the path defined in the dataset
        path: Optional[str] = None,
        # The compression type to use for any storage format that allows compression to be specified.
        write_compression: Optional[str] = None,
        # A list of columns by which the CTAS table will be partitioned.
        partitioning_info: Optional[List[str]] = None,
        # Tuple consisting of the column names used for bucketing as the
        # first element and the number of buckets as the second element.
        # Only str, int and bool are supported as column data types for bucketing.
        bucketing_info: Optional[Tuple[List[str], int]] = None,
        # Athena workgroup.
        workgroup: Optional[str] = None,
        # Data Source / Catalog name. If None, ‘AwsDataCatalog’ is used.
        data_source: Optional[str] = None,
        # Whether to wait for the query to finish and return a dictionary with the Query metadata.
        wait: bool = False,
        # Boto3 Session. The default boto3 Session will be used if boto3_session receive None.,
        boto3_session: Optional[Any] = None,
        # Drop the table before running CTAS query
        drop_before_create: bool = False,
    ) -> bool:
        """
        Create dataset from query
        """

        # S3Dataset not yet initialized
        if self.args is None:
            return False

        if sql is None:
            logger.error("Sql not provided")
            return False

        dataset_path = path or self.uri
        if dataset_path is None:
            logger.error("Dataset path not provided")
            return False
        logger.info(f"Writing to {dataset_path}")

        database = origin_database or self.database
        if database is None:
            logger.error("Database not provided")
            return False

        # create a dict of args which are not null
        not_null_args: Dict[str, Any] = {}

        _write_compression = write_compression or self.compression
        if _write_compression is not None:
            not_null_args["write_compression"] = _write_compression

        _partitioning_info = partitioning_info or self.partition_cols
        if _partitioning_info is not None:
            not_null_args["partitioning_info"] = _partitioning_info

        _bucketing_info = bucketing_info or self.bucketing_info
        if _bucketing_info is not None:
            not_null_args["bucketing_info"] = _bucketing_info

        if workgroup is not None:
            not_null_args["workgroup"] = workgroup

        if data_source is not None:
            not_null_args["data_source"] = data_source

        _boto3_session = boto3_session or self.boto3_session
        if _boto3_session is not None:
            not_null_args["boto3_session"] = _boto3_session

        try:
            import awswrangler as wr

            if drop_before_create:
                self._delete(path=dataset_path, boto3_session=_boto3_session)

            response: Dict[str, Union[str, Any]] = wr.athena.create_ctas_table(
                sql=sql,
                database=database,
                ctas_table=self.table,
                ctas_database=self.database,
                s3_output=dataset_path,
                storage_format="PARQUET",
                write_compression=_write_compression,
                wait=wait,
                **not_null_args,
            )
            logger.info(f"Dataset {self.name} written to {dataset_path}")
            logger.debug(f"Response: {response}")
            logger.info("--**-- Done")
            return True
        except Exception as e:
            logger.error("Could not write to dataset: {}".format(self.name))
            raise

    ######################################################
    ## Delete DataAsset
    ######################################################

    def _delete(
        self,
        # S3 path for the dataset
        path: Optional[str] = None,
        # Boto3 Session. The default boto3 Session will be used if boto3_session receive None.
        boto3_session: Optional[Any] = None,
    ) -> bool:
        import awswrangler as wr

        table_exists = wr.catalog.does_table_exist(
            table=self.table,
            database=self.database,
            catalog_id=self.catalog_id,
            boto3_session=boto3_session,
        )
        if not table_exists:
            return True

        s3_path = path or self.uri
        logger.info(f"Deleting objects at {s3_path}")
        wr.s3.delete_objects(s3_path)

        logger.info(f"Dropping table {self.name}")
        drop_table = wr.catalog.delete_table_if_exists(
            database=self.database,
            table=self.table,
            catalog_id=self.catalog_id,
            boto3_session=boto3_session,
        )
        if drop_table:
            logger.info(f"Table {self.name} dropped")
        return drop_table



from datetime import datetime
from pathlib import Path
from typing import Optional

from phidata.constants import (
    SCRIPTS_DIR_ENV_VAR,
    STORAGE_DIR_ENV_VAR,
    META_DIR_ENV_VAR,
    PRODUCTS_DIR_ENV_VAR,
    NOTEBOOKS_DIR_ENV_VAR,
    WORKSPACE_CONFIG_DIR_ENV_VAR,
)
from phidata.types.context import PathContext, RunContext, AirflowContext
from phidata.utils.log import logger


def get_run_date(
    run_context: Optional[RunContext] = None,
    airflow_context: Optional[AirflowContext] = None,
) -> datetime:

    if run_context is not None and run_context.run_date is not None:
        return run_context.run_date
    if airflow_context is not None and airflow_context.logical_date is not None:
        return airflow_context.logical_date
    return datetime.now()


def build_path_context_from_env() -> Optional[PathContext]:

    logger.debug(f"--++**++--> Building PathContext from env")
    path_context = PathContext()

    import os

    scripts_dir = os.getenv(SCRIPTS_DIR_ENV_VAR)
    storage_dir = os.getenv(STORAGE_DIR_ENV_VAR)
    meta_dir = os.getenv(META_DIR_ENV_VAR)
    products_dir = os.getenv(PRODUCTS_DIR_ENV_VAR)
    notebooks_dir = os.getenv(NOTEBOOKS_DIR_ENV_VAR)
    workspace_config_dir = os.getenv(WORKSPACE_CONFIG_DIR_ENV_VAR)

    if storage_dir is None:
        logger.error(f"{STORAGE_DIR_ENV_VAR} not set")
    if products_dir is None:
        logger.error(f"{PRODUCTS_DIR_ENV_VAR} not set")

    try:
        if scripts_dir is not None:
            path_context.scripts_dir = Path(scripts_dir)
        if storage_dir is not None:
            path_context.storage_dir = Path(storage_dir)
        if meta_dir is not None:
            path_context.meta_dir = Path(meta_dir)
        if products_dir is not None:
            path_context.products_dir = Path(products_dir)
        if notebooks_dir is not None:
            path_context.notebooks_dir = Path(notebooks_dir)
        if workspace_config_dir is not None:
            path_context.workspace_config_dir = Path(workspace_config_dir)
    except Exception as e:
        raise
    logger.debug(f"--++**++--> PathContext loaded")
    return path_context
